# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import bisect
import collections
import math

from rally.common.plugin import plugin
from rally.common import streaming_algorithms as streaming
from rally.task import atomic
from rally.task.processing import utils


@plugin.base()
class Chart(plugin.Plugin, metaclass=abc.ABCMeta):
    """Base class for charts.

    This is a base for all plugins that prepare data for specific charts
    in HTML report. Each chart must at least declare chart widget and
    prepare data that is suitable for rendering by JavaScript.
    """

    @abc.abstractproperty
    def widget(self):
        """Widget name to display this chart by JavaScript."""

    def __init__(self, workload, zipped_size=1000):
        """Setup initial values.

        :param workload: dict, detailed info about the Workload
        :param zipped_size: int maximum number of points on scale
        """
        self._data = collections.OrderedDict()  # Container for results
        self._workload = workload
        self.base_size = self._workload["total_iteration_count"]
        self.zipped_size = zipped_size

    def add_iteration(self, iteration):
        """Add iteration data.

        This method must be called for each iteration.
        If overridden, this method must use streaming data processing,
        so chart instance could process unlimited number of iterations,
        with low memory usage.
        """
        for name, value in self._map_iteration_values(iteration):
            if name not in self._data:
                self._data[name] = utils.GraphZipper(self.base_size,
                                                     self.zipped_size)
            self._data[name].add_point(value)

    def render(self):
        """Generate chart data ready for drawing."""
        return [(name, points.get_zipped_graph())
                for name, points in self._data.items()]

    @classmethod
    def render_complete_data(cls, data):
        """render processed complete data for drawing."""
        return data

    def _fix_atomic_actions(self, atomic_actions):
        """Set `0' for missed atomic actions.

        Since some atomic actions can absent in some iterations
        due to failures, this method must be used in all cases
        related to atomic actions processing.
        """
        return list(
            (name, atomic_actions.get(name, {}).get("duration", 0))
            for name in self._get_atomic_names()
        )

    def _get_atomic_names(self):
        duration_stats = self._workload["statistics"]["durations"]
        return [a["display_name"] for a in duration_stats["atomics"]]

    def _map_iteration_values(self, iteration):
        """Get values for processing, from given iteration."""
        return iteration


class MainStackedAreaChart(Chart):

    widget = "StackedArea"

    def _map_iteration_values(self, iteration):
        if iteration["error"]:
            result = [("duration", 0), ("idle_duration", 0)]
            if self._workload["failed_iteration_count"]:
                result.append(
                    ("failed_duration",
                     iteration["duration"] + iteration["idle_duration"]))
        else:
            result = [("duration", iteration["duration"]),
                      ("idle_duration", iteration["idle_duration"])]
            if self._workload["failed_iteration_count"]:
                result.append(("failed_duration", 0))
        return result


class AtomicStackedAreaChart(Chart):

    widget = "StackedArea"

    def _map_iteration_values(self, iteration):
        atomic_actions = atomic.merge_atomic_actions(
            iteration["atomic_actions"])
        atomics = self._fix_atomic_actions(atomic_actions)
        if self._workload["failed_iteration_count"]:
            if iteration["error"]:
                failed_duration = (
                    iteration["duration"] + iteration["idle_duration"]
                    - sum([(a[1] or 0) for a in atomics]))
            else:
                failed_duration = 0
            atomics.append(("failed_duration", failed_duration))
        return atomics


class AvgChart(Chart):
    """Base class for charts with average results."""

    widget = "Pie"

    def add_iteration(self, iteration):
        for name, value in self._map_iteration_values(iteration):
            if name not in self._data:
                self._data[name] = streaming.MeanComputation()
            self._data[name].add(value or 0)

    def render(self):
        return [(k, v.result()) for k, v in self._data.items()]


class AtomicAvgChart(AvgChart):

    def _map_iteration_values(self, iteration):
        atomic_actions = atomic.merge_atomic_actions(
            iteration["atomic_actions"])
        return self._fix_atomic_actions(atomic_actions)


class LoadProfileChart(Chart):
    """Chart for parallel durations."""

    widget = "StackedArea"

    def __init__(self, workload, name="parallel iterations",
                 scale=100):
        """Setup chart with graph name and scale.

        :param workload:  dict, detailed information about Workload
        :param name: str name for X axis
        :param scale: int number of X points
        """
        super(LoadProfileChart, self).__init__(workload)
        self._name = name
        # NOTE(boris-42): Add 2 points at the end of graph so at the end of
        #                 graph there will be point with 0 running iterations.
        self._duration = self._workload["load_duration"] * (1 + 2.0 / scale)

        self.step = self._duration / float(scale)
        self._time_axis = [self.step * x
                           for x in range(int(scale))
                           if (self.step * x) < self._duration]
        self._time_axis.append(self._duration)
        self._running = [0] * len(self._time_axis)
        # NOTE(andreykurilin): There is a "start_time" field in workload
        #   object, but due to transformations in database layer, the
        #   microseconds can be not accurate enough.
        if self._workload["data"]:
            self._tstamp_start = self._workload["data"][0]["timestamp"]
        else:
            self._tstamp_start = self._workload["start_time"]

    def _map_iteration_values(self, iteration):
        return iteration["timestamp"], iteration["duration"]

    def add_iteration(self, iteration):
        timestamp, duration = self._map_iteration_values(iteration)
        ts_start = timestamp - self._tstamp_start
        started_idx = bisect.bisect(self._time_axis, ts_start)
        ended_idx = bisect.bisect(self._time_axis, ts_start + duration)
        if self._time_axis[ended_idx - 1] == ts_start + duration:
            ended_idx -= 1
        for idx in range(started_idx + 1, ended_idx):
            self._running[idx] += 1
        if started_idx == ended_idx:
            self._running[ended_idx] += duration / self.step
        else:
            self._running[started_idx] += (
                self._time_axis[started_idx] - ts_start) / self.step
            self._running[ended_idx] += (
                ts_start + duration
                - self._time_axis[ended_idx - 1]) / self.step

    def render(self):
        return [(self._name, list(zip(self._time_axis, self._running)))]


class HistogramChart(Chart):
    """Base class for chart with histograms.

    This chart is relatively complex, because actually it is a set
    of histograms, that usually can be switched by dropdown select.
    And each histogram has several data views.
    """

    widget = "Histogram"

    def _init_views(self, min_value, max_value):
        """Generate initial data for each histogram view."""
        if not self.base_size:
            return []
        min_value, max_value = min_value or 0, max_value or 0
        views = []
        for view, bins in [
                ("Square Root Choice",
                 int(math.ceil(math.sqrt(self.base_size)))),
                ("Sturges Formula",
                 int(math.ceil(math.log(self.base_size, 2) + 1))),
                ("Rice Rule",
                 int(math.ceil(2 * self.base_size ** (1.0 / 3))))]:
            bin_width = float(max_value - min_value) / bins
            x_axis = [min_value + (bin_width * x) for x in range(1, bins + 1)]
            views.append({"view": view, "bins": bins,
                          "x": x_axis, "y": [0] * len(x_axis)})
        return views

    def add_iteration(self, iteration):
        for name, value in self._map_iteration_values(iteration):
            if name not in self._data:
                raise KeyError("Unexpected histogram name: %s" % name)
            for i, view in enumerate(self._data[name]["views"]):
                for bin_i, bin_v in enumerate(view["x"]):
                    if (value or 0) <= bin_v:
                        self._data[name]["views"][i]["y"][bin_i] += 1
                        break

    def render(self):
        data = []
        for name, hist in self._data.items():
            for idx, v in enumerate(hist["views"]):
                graph = {"key": name,
                         "view": v["view"],
                         "disabled": hist["disabled"],
                         "values": [{"x": x, "y": y}
                                    for x, y in zip(v["x"], v["y"])]}
                try:
                    data[idx].append(graph)
                except IndexError:
                    data.append([graph])
        return {"data": data, "views": [{"id": i, "name": d[0]["view"]}
                                        for i, d in enumerate(data)]}


class MainHistogramChart(HistogramChart):

    def __init__(self, workload_info):
        super(MainHistogramChart, self).__init__(workload_info)
        views = self._init_views(self._workload["min_duration"],
                                 self._workload["max_duration"])
        self._data["task"] = {"views": views, "disabled": None}

    def _map_iteration_values(self, iteration):
        return [("task", 0 if iteration["error"] else iteration["duration"])]


class AtomicHistogramChart(HistogramChart):

    def __init__(self, workload_info):
        super(AtomicHistogramChart, self).__init__(workload_info)
        for i, aa in enumerate(
                self._workload["statistics"]["durations"]["atomics"]):
            self._data[aa["display_name"]] = {
                "views": self._init_views(aa["data"]["min"],
                                          aa["data"]["max"]),
                "disabled": i}

    def _map_iteration_values(self, iteration):
        atomic_actions = atomic.merge_atomic_actions(
            iteration["atomic_actions"])
        return self._fix_atomic_actions(atomic_actions)


class Table(Chart, metaclass=abc.ABCMeta):
    """Base class for tables.

    Each Table subclass represents HTML table which can be easily rendered in
    report. Subclasses are responsible for setting up both columns and rows:
    columns are set simply by `columns' property (list of str columns names)
    and rows must be initialized in _data property, with the following format:
        self._data = {name: [streaming_ins, postprocess_func or None], ...}
          where:
            name - str name of table row parameter
            streaming_ins - instance of streaming algorithm
            postprocess_func - optional function that processes final result,
                               None means usage of default self._round()
        This can be done in __init__() or even in add_iteration().
    """

    widget = "Table"
    _styles = {}

    @abc.abstractproperty
    def columns(self):
        """List of columns names."""

    def _round(self, ins, has_result):
        """This is a default post-process function for table cell value.

        :param ins: streaming_algorithms.StreamingAlgorithm subclass instance
        :param has_result: bool, whether current row is effective
        :returns: rounded float
        :returns: str "n/a"
        """
        if isinstance(ins, streaming.StreamingAlgorithm):
            ins = ins.result()

        if not has_result or ins is None:
            return "n/a"
        else:
            return round(ins, 3)

    def _row_has_results(self, values):
        """Determine whether row can be assumed as having values.

        :param values: row values list
                       [(StreamingAlgorithm, function or None), ...]
        :returns: bool
        """
        for ins, fn in values:
            if isinstance(ins, (streaming.MinComputation,
                                streaming.MaxComputation,
                                streaming.MeanComputation)):
                # NOTE(amaretskiy): None means this computation
                #                   has never been called
                return ins.result() is not None
        return True

    def _process_row(self, name, values):
        row = [name]
        has_result = self._row_has_results(values)
        for ins, fn in values:
            fn = fn or self._round
            row.append(fn(ins, has_result))
        return row

    def get_rows(self):
        """Collect rows values finally, after all data is processed.

        :returns: [str_name, (float or str), (float or str), ...]
        """
        rows = []
        for name, values in self._data.items():
            rows.append(self._process_row(name, values))
        return rows

    def render(self):
        rows = self.get_rows()
        if self._styles is None:
            # do not apply anything
            styles = {}
        elif not self._styles and rows:
            # make the last elements bold
            styles = {len(rows) - 1: "rich"}
        else:
            styles = self._styles
        return {"cols": self.columns,
                "rows": rows,
                "styles": styles}


class MainStatsTable(Table):

    columns = ["Action", "Min (sec)", "Median (sec)", "90%ile (sec)",
               "95%ile (sec)", "Max (sec)", "Avg (sec)", "Success", "Count"]

    _DEPTH_OF_PROCESSING = 2

    def __init__(self, *args, **kwargs):
        super(MainStatsTable, self).__init__(*args, **kwargs)
        self.iters_num = self._workload["total_iteration_count"]

    def _initialize_atomic(self, name, root, real_name=None, count=1):
        real_name = real_name or name
        root[name] = {
            # streaming algorithms
            "sa": [
                streaming.PointsSaver(),
                streaming.MinComputation(),
                streaming.MaxComputation(),
                streaming.MeanComputation(),
                streaming.MeanComputation(),
                streaming.IncrementComputation()
            ],
            "children": collections.OrderedDict(),
            "real_name": real_name,
            "count_per_iteration": count
        }

    def _add_data(self, raw_data, root=None):
        """Add iteration data."""
        p_data = self._data if root is None else root
        for name, data in raw_data.items():
            original_name = name
            if data["count"] > 1:
                name += (" (x%s)" % data["count"])
            if name not in p_data:
                self._initialize_atomic(name,
                                        root=p_data,
                                        real_name=original_name,
                                        count=data["count"])

            points, min_v, max_v, mean, success, count = p_data[name]["sa"]
            count.add()
            success.add(0 if data.get("failed", False) else 1)
            for sa in (points, min_v, max_v, mean):
                sa.add(data["duration"])

            if data["children"]:
                self._add_data(data["children"], root=p_data[name]["children"])

    def add_iteration(self, iteration):
        """Add data of a single iteration."""
        data = atomic.merge_atomic_actions(iteration["atomic_actions"])
        # NOTE(andreykurilin): the easiest way to identify the last
        #   atomic is to find the last added key to the OrderedDict. The
        #   most perfect way is to use reversed, since class OrderedDict
        #   uses a doubly linked list for the dictionary items and
        #   implements __reversed__(), what is why such implementation
        #   gives you O(1) access to the desired element.
        if data:
            the_last = data[next(reversed(data))]
            if iteration["error"] and not the_last.get("failed", False):
                # un-wrapped action failed
                data["<no-name-action>"] = {"duration": 0, "count": 1,
                                            "failed": True, "children": {}}
        total_duration = iteration["duration"] + iteration["idle_duration"]
        data["total"] = {"duration": total_duration,
                         "count": 1,
                         "failed": bool(iteration["error"]),
                         "children": collections.OrderedDict(
                             [("duration", {
                                 "duration": iteration["duration"],
                                 "count": 1,
                                 "failed": bool(iteration["error"]),
                                 "children": []}),
                              ("idle_duration", {
                                  "duration": iteration["idle_duration"],
                                  "count": 1,
                                  "failed": bool(iteration["error"]),
                                  "children": []})
                              ])}

        self._add_data(data)

    def _process_row(self, sa):
        points_ins, min_v, max_v, avg, success, count = sa

        # process percentiles
        points = points_ins.result()
        # clear the space on the disk
        points_ins.reset()
        # sort data once
        points.sort()
        p50ile = utils.percentile(points, 0.5, ignore_sorting=True)
        p90ile = utils.percentile(points, 0.9, ignore_sorting=True)
        p95ile = utils.percentile(points, 0.95, ignore_sorting=True)

        # process and round values
        count = count.result()
        has_result = bool(count)
        min_v = self._round(min_v, has_result)
        max_v = self._round(max_v, has_result)
        avg = self._round(avg, has_result)
        p50ile = self._round(p50ile, has_result)
        p90ile = self._round(p90ile, has_result)
        p95ile = self._round(p95ile, has_result)
        success = "%.1f%%" % (success.result() * 100) if has_result else "n/a"

        return min_v, max_v, avg, p50ile, p90ile, p95ile, success, count

    def _process_result(self, name, values):
        r = self._process_row(values["sa"])
        min_v, max_v, avg, p50ile, p90ile, p95ile, success, count = r
        children = []

        for c_name, c_values in values["children"].items():
            children.append(self._process_result(c_name, c_values))
        return {"data": {"iteration_count": count,
                         "min": min_v,
                         "median": p50ile,
                         "90%ile": p90ile,
                         "95%ile": p95ile,
                         "max": max_v,
                         "avg": avg,
                         "success": success},
                "count_per_iteration": values["count_per_iteration"],
                "name": values["real_name"],
                "display_name": name,
                "children": children}

    def _get_results(self):
        if self._data:
            # NOTE(andreykurilin): In case when the specific atomic action was
            #   not executed in the first iteration, it will be added to
            #   self._data after the "total" raw. It is a wrong order, so
            #   let's ensure that the "total" is always at the end
            self._data["total"] = self._data.pop("total")
        else:
            # NOTE(andreykurilin): The workload doesn't have any iteration, so
            #   the method `add_iteration` had not been called and 'total' item
            #   had not be initialized. Let's do it here, since 'total' item
            #   should always present in the rows.
            self._initialize_atomic("total", root=self._data)

        results = []
        for name, values in self._data.items():
            results.append(self._process_result(name, values))
        return results

    def get_rows(self):
        rows = []

        def _process_elem(elem, depth=0):
            name = elem["display_name"]
            if depth > 0:
                name = (" %s> %s" % ("-" * depth, name))
            rows.append([name,
                         elem["data"]["min"],
                         elem["data"]["median"],
                         elem["data"]["90%ile"],
                         elem["data"]["95%ile"],
                         elem["data"]["max"],
                         elem["data"]["avg"],
                         elem["data"]["success"],
                         elem["data"]["iteration_count"]])
            for child in elem["children"]:
                _process_elem(child, depth=(depth + 1))

        for elem in self._get_results():
            _process_elem(elem)

        return rows

    def to_dict(self):
        res = self._get_results()
        return {"total": res[-1], "atomics": res[:-1]}

    def render(self):
        rendered_data = super(MainStatsTable, self).render()
        rows_len = len(rendered_data["rows"])
        if rows_len > 1:
            styles = {rows_len - 3: "rich",
                      rows_len - 2: "oblique",
                      rows_len - 1: "oblique"}
            for i, row in enumerate(rendered_data["rows"]):
                if i == rows_len - 3:
                    break
                if row[0].startswith(" -"):
                    styles[i] = "oblique"
            rendered_data["styles"] = styles
        return rendered_data


class OutputChart(Chart):
    """Base class for charts related to scenario output."""

    def __init__(self, workload_info, zipped_size=1000,
                 title="", description="", label="", axis_label=""):
        super(OutputChart, self).__init__(workload_info, zipped_size)
        self.title = title
        self.description = description
        self.label = label
        self.axis_label = axis_label

    def render(self):
        return {"title": self.title,
                "description": self.description,
                "widget": self.widget,
                "data": super(OutputChart, self).render(),
                "label": self.label,
                "axis_label": self.axis_label}


@plugin.configure(name="StackedArea")
class OutputStackedAreaChart(OutputChart):
    """Display results as stacked area.

    This plugin processes additive data and displays it in HTML report
    as stacked area with X axis bound to iteration number.
    Complete output data is displayed as stacked area as well, without
    any processing.

    Keys "description", "label" and "axis_label" are optional.

    Examples of using this plugin in Scenario, for saving output data:

    .. code-block:: python

        self.add_output(
            additive={"title": "Additive data as stacked area",
                      "description": "Iterations trend for foo and bar",
                      "chart_plugin": "StackedArea",
                      "data": [["foo", 12], ["bar", 34]]},
            complete={"title": "Complete data as stacked area",
                      "description": "Data is shown as stacked area, as-is",
                      "chart_plugin": "StackedArea",
                      "data": [["foo", [[0, 5], [1, 42], [2, 15], [3, 7]]],
                               ["bar", [[0, 2], [1, 1.3], [2, 5], [3, 9]]]],
                      "label": "Y-axis label text",
                      "axis_label": "X-axis label text"})
    """

    widget = "StackedArea"

    def render(self):
        result = super(OutputStackedAreaChart, self).render()

        # NOTE(amaretskiy): transform to Table if there is a single iteration
        if result["data"] and len(result["data"][0][1]) == 1:
            rows = [[v[0], v[1][0][1]] for v in result["data"]]
            result.update({"widget": "Table",
                           "data": {"cols": ["Name", self.label or "Value"],
                                    "rows": rows}})
        return result


@plugin.configure(name="Lines")
class OutputLinesChart(OutputStackedAreaChart):
    """Display results as generic chart with lines.

    This plugin processes additive data and displays it in HTML report
    as linear chart with X axis bound to iteration number.
    Complete output data is displayed as linear chart as well, without
    any processing.

    Examples of using this plugin in Scenario, for saving output data:

    .. code-block:: python

        self.add_output(
            additive={"title": "Additive data as stacked area",
                      "description": "Iterations trend for foo and bar",
                      "chart_plugin": "Lines",
                      "data": [["foo", 12], ["bar", 34]]},
            complete={"title": "Complete data as stacked area",
                      "description": "Data is shown as stacked area, as-is",
                      "chart_plugin": "Lines",
                      "data": [["foo", [[0, 5], [1, 42], [2, 15], [3, 7]]],
                               ["bar", [[0, 2], [1, 1.3], [2, 5], [3, 9]]]],
                      "label": "Y-axis label text",
                      "axis_label": "X-axis label text"})
    """

    widget = "Lines"


@plugin.configure(name="Pie")
class OutputAvgChart(OutputChart, AvgChart):
    """Display results as pie, calculate average values for additive data.

    This plugin processes additive data and calculate average values.
    Both additive and complete data are displayed in HTML report as pie chart.

    Examples of using this plugin in Scenario, for saving output data:

    .. code-block:: python

        self.add_output(
            additive={"title": "Additive output",
                      "description": ("Pie with average data "
                                      "from all iterations values"),
                      "chart_plugin": "Pie",
                      "data": [["foo", 12], ["bar", 34], ["spam", 56]]},
            complete={"title": "Complete output",
                      "description": "Displayed as a pie, as-is",
                      "chart_plugin": "Pie",
                      "data": [["foo", 12], ["bar", 34], ["spam", 56]]})
    """

    widget = "Pie"


@plugin.configure(name="Table")
class OutputTable(OutputChart, Table):
    """Display complete output as table, can not be used for additive data.

    Use this plugin for complete output data to display it in HTML report
    as table. This plugin can not be used for additive data because it
    does not contain any processing logic.

    Examples of using this plugin in Scenario, for saving output data:

    .. code-block:: python

        self.add_output(
            complete={"title": "Arbitrary Table",
                      "description": "Just show columns and rows as-is",
                      "chart_plugin": "Table",
                      "data": {"cols": ["foo", "bar", "spam"],
                               "rows": [["a row", 1, 2], ["b row", 3, 4],
                                        ["c row", 5, 6]]}})
    """

    widget = "Table"


@plugin.configure(name="StatsTable")
class OutputStatsTable(OutputTable):
    """Calculate statistics for additive data and display it as table.

    This plugin processes additive data and compose statistics that is
    displayed as table in HTML report.

    Examples of using this plugin in Scenario, for saving output data:

    .. code-block:: python

        self.add_output(
            additive={"title": "Statistics",
                      "description": ("Table with statistics generated "
                                      "from all iterations values"),
                      "chart_plugin": "StatsTable",
                      "data": [["foo stat", 12], ["bar", 34], ["spam", 56]]})
    """

    columns = ["Action", "Min (sec)", "Median (sec)", "90%ile (sec)",
               "95%ile (sec)", "Max (sec)", "Avg (sec)", "Count"]

    def add_iteration(self, iteration):
        for name, value in self._map_iteration_values(iteration):
            if name not in self._data:
                self._data[name] = [
                    streaming.PointsSaver(),
                    streaming.IncrementComputation(),
                    streaming.MinComputation(),
                    streaming.MaxComputation(),
                    streaming.MeanComputation()
                ]
            points, count, min_v, max_v, avg = self._data[name]

            count.add()
            for ins in (points, min_v, max_v, avg):
                ins.add(value)

    def _process_row(self, name, sa):
        points_ins, count, min_v, max_v, avg = sa

        # process percentiles
        points = points_ins.result()
        # clear the space on the disk
        points_ins.reset()
        # sort data once
        points.sort()
        p50ile = utils.percentile(points, 0.5, ignore_sorting=True)
        p90ile = utils.percentile(points, 0.9, ignore_sorting=True)
        p95ile = utils.percentile(points, 0.95, ignore_sorting=True)

        # process and round values
        count = count.result()
        has_result = bool(count)

        min_v = self._round(min_v, has_result)
        max_v = self._round(max_v, has_result)
        avg = self._round(avg, has_result)
        p50ile = self._round(p50ile, has_result)
        p90ile = self._round(p90ile, has_result)
        p95ile = self._round(p95ile, has_result)

        return [name, min_v, p50ile, p90ile, p95ile, max_v, avg, count]

    def get_rows(self):
        """Collect rows values finally, after all data is processed.

        :returns: [str_name, (float or str), (float or str), ...]
        """
        rows = []
        for name, values in self._data.items():
            rows.append(self._process_row(name, values))
        return rows


@plugin.configure(name="TextArea")
class OutputTextArea(OutputChart):
    """Arbitrary text

    This plugin processes complete data and displays of output in HTML report.

    Examples of using this plugin in Scenario, for saving output data:

    .. code-block:: python

        self.add_output(
            complete={"title": "Script Inline",
                      "chart_plugin": "TextArea",
                      "data": ["first output", "second output",
                               "third output"]]})
    """

    widget = "TextArea"


@plugin.configure(name="EmbeddedChart")
class OutputEmbeddedChart(OutputChart):
    """Chart for embedding custom html as a complete chart.

    Example of usage:

    .. code-block:: python

        self.add_output(
            complete={
                "title": "Embedding link to example.com",
                "chart_plugin": "EmbeddedChart",
                "data": "<a href='example.com'>"
                        "To see external logs follow this link"
                        "</a>"
            }
        )
    """

    widget = "EmbedChart"

    @classmethod
    def render_complete_data(cls, pdata):
        return {
            "title": pdata["title"],
            "widget": cls.widget,
            "data": {
                "source": None,
                # NOTE(chenxu): ensure that '</script>' of embedded_data will
                #   not be handled incorrectly by JavaScript.
                "embedded": pdata["data"].replace("/script>", "\\/script>"),
            }
        }


@plugin.configure(name="EmbeddedExternalChart")
class OutputEmbeddedExternalChart(OutputChart):
    """Chart for embedding external html page as a complete chart.

    Example of usage:

    .. code-block:: python

        self.add_output(
            complete={
                "title": "Embedding external html page",
                "chart_plugin": "EmbeddedExternalChart",
                "data": "https://example.com"
            }
        )
    """

    widget = "EmbedChart"

    @classmethod
    def render_complete_data(cls, pdata):
        return {
            "title": pdata["title"],
            "widget": cls.widget,
            "data": {"embedded": None, "source": pdata["data"]}
        }


_OUTPUT_SCHEMA = {
    "key_types": {
        "title": str,
        "description": str,
        "chart_plugin": str,
        "data": (list, dict),
        "label": str,
        "axis_label": str},
    "required": ["title", "chart_plugin", "data"]}


def validate_output(output_type, output):
    # TODO(amaretskiy): this validation is simple and must be improved.
    #   Maybe it is worth to add classmethod OutputChart.validate(), so
    #   we could have flexible validation for custom chart plugins
    if output_type not in ("additive", "complete"):
        return ("unexpected output type: '%s', "
                "should be in ('additive', 'complete')" % output_type)

    if not isinstance(output, dict):
        return ("%(name)s output item has wrong type '%(type)s', "
                "must be 'dict'" % {"name": output_type,
                                    "type": type(output).__name__})

    for key in _OUTPUT_SCHEMA["required"]:
        if key not in output:
            return ("%(name)s output missing key '%(key)s'"
                    % {"name": output_type, "key": key})

    for key in output:
        if key not in _OUTPUT_SCHEMA["key_types"]:
            return ("%(name)s output has unexpected key '%(key)s'"
                    % {"name": output_type, "key": key})

        proper_type = _OUTPUT_SCHEMA["key_types"][key]
        if not isinstance(output[key], proper_type):
            if type(proper_type) == tuple:
                return ("Value of %(name)s output %(key)s has wrong type "
                        "'%(actual_type)s', should be in %(types)r"
                        % {"name": output_type,
                           "key": key,
                           "actual_type": type(output[key]).__name__,
                           "types": tuple(t.__name__
                                          for t in proper_type)})
            return ("Value of %(name)s output %(key)s has wrong type "
                    "'%(actual_type)s', should be %(proper_type)s"
                    % {"name": output_type,
                       "key": key,
                       "actual_type": type(output[key]).__name__,
                       "proper_type": proper_type.__name__})



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import json

from rally.common.plugin import plugin
from rally import plugins
from tests.unit import test


class ConfigSchemasTestCase(test.TestCase):

    OBJECT_TYPE_KEYS = {"$schema", "type", "description", "required",
                        "properties", "patternProperties",
                        "minProperties", "maxProperties",
                        "additionalProperties", "oneOf", "anyOf"}
    ARRAY_TYPE_KEYS = {"$schema", "type", "description", "items",
                       "uniqueItems", "minItems", "maxItems",
                       "additionalItems"}
    NUMBER_TYPE_KEYS = {"$schema", "type", "description", "minimum", "maximum",
                        "exclusiveMinimum"}
    STRING_TYPE_KEYS = {"$schema", "type", "description", "pattern"}

    def fail(self, p, schema, msg):
        super(ConfigSchemasTestCase, self).fail(
            "Config schema of plugin '%s' (%s) is invalid. %s "
            "Schema: \n%s" % (p.get_name(),
                              "%s.%s" % (p.__module__, p.__name__),
                              msg,
                              json.dumps(schema, indent=3)))

    def _check_anyOf_or_oneOf(self, p, schema, definitions):
        if "anyOf" in schema or "oneOf" in schema:
            key = "anyOf" if "anyOf" in schema else "oneOf"
            for case in schema[key]:
                if "description" not in case and "$ref" not in case:
                    self.fail(p, schema, "Each case of '%s' should have "
                                         "description." % key)
                full_schema = copy.deepcopy(schema)
                full_schema.pop(key)
                for k, v in case.items():
                    full_schema[k] = v
                self._check_item(p, full_schema, definitions)

    def _check_object_type(self, p, schema, definitions):
        unexpected_keys = set(schema.keys()) - self.OBJECT_TYPE_KEYS
        if "definitions" in unexpected_keys:
            # TODO(andreykurilin): do not use definitions since it is a hard
            #     task to parse and display them
            unexpected_keys -= {"definitions"}
        if unexpected_keys:
            self.fail(p, schema, ("Found unexpected key(s) for object type: "
                                  "%s." % ", ".join(unexpected_keys)))
        if "additionalProperties" not in schema:
            self.fail(p, schema,
                      "'additionalProperties' is required field for objects. "
                      "Specify `'additionalProperties': True` explicitly to "
                      "accept not validated properties.")

        if "patternProperties" in schema:
            if "properties" in schema:
                self.fail(p, schema, "Usage both 'patternProperties' and "
                                     "'properties' in one time is restricted.")
            if not isinstance(schema["patternProperties"], dict):
                self.fail(p, schema, "Field 'patternProperties' should be a "
                                     "dict.")
            for pattern, description in schema["patternProperties"].items():
                self._check_item(p, description, definitions)

        if "properties" in schema:
            for property_name, description in schema["properties"].items():
                self._check_item(p, description, definitions)

    def _check_array_type(self, p, schema, definitions):
        unexpected_keys = set(schema.keys()) - self.ARRAY_TYPE_KEYS
        if "additionalProperties" in unexpected_keys:
            self.fail(p, schema, "Array type doesn't support "
                                 "'additionalProperties' field.")

        if unexpected_keys:
            self.fail(p, schema, ("Found unexpected key(s) for array type: "
                                  "%s." % ", ".join(unexpected_keys)))

        if "items" not in schema:
            self.fail(p, schema, "Expected items of array type should be "
                                 "described via 'items' field.")

        if isinstance(schema["items"], dict):
            self._check_item(p, schema["items"], definitions)
            if "additionalItems" in schema:
                self.fail(p, schema, "When items is a single schema, the "
                                     "`additionalItems` keyword is "
                                     "meaningless, and it should not be used.")
        elif isinstance(schema["items"], list):
            for item in schema["items"]:
                self._check_item(p, item, definitions)
        else:
            self.fail(p, schema, ("Field 'items' of array type should be a "
                                  "list or a dict, but not '%s'" %
                                  type(schema["items"])))

    def _check_string_type(self, p, schema):
        unexpected_keys = set(schema.keys()) - self.STRING_TYPE_KEYS
        if unexpected_keys:
            self.fail(p, schema, ("Found unexpected key(s) for string type: "
                                  "%s." % ", ".join(unexpected_keys)))

    def _check_number_type(self, p, schema):
        unexpected_keys = set(schema.keys()) - self.NUMBER_TYPE_KEYS
        if unexpected_keys:
            self.fail(p, schema, ("Found unexpected key(s) for integer/number "
                                  "type: %s." % ", ".join(unexpected_keys)))

    def _check_simpliest_types(self, p, schema, type_name):
        unexpected_keys = set(schema.keys()) - {"type", "description"}
        if unexpected_keys:
            self.fail(p, schema, ("Found unexpected key(s) for %s type: "
                                  "%s." % (type_name,
                                           ", ".join(unexpected_keys))))

    def _check_item(self, p, schema, definitions):
        if "type" in schema or "anyOf" in schema or "oneOf" in schema:
            if "anyOf" in schema or "oneOf" in schema:
                self._check_anyOf_or_oneOf(p, schema, definitions)
            elif "type" in schema:
                if schema["type"] == "object":
                    self._check_object_type(p, schema, definitions)
                elif schema["type"] == "array":
                    self._check_array_type(p, schema, definitions)
                elif schema["type"] == "string":
                    self._check_string_type(p, schema)
                elif schema["type"] in ("number", "integer"):
                    self._check_number_type(p, schema)
                elif schema["type"] in ("boolean", "null"):
                    self._check_simpliest_types(p, schema, schema["type"])
                elif isinstance(schema["type"], list):
                    self._check_simpliest_types(p, schema, "mixed")
                else:
                    self.fail(p, schema,
                              "Wrong type is used: %s" % schema["type"])
        elif "enum" in schema:
            pass
        elif schema == {}:
            # NOTE(andreykurilin): an empty dict means that the user can
            #   transmit whatever he wants in whatever he wants format. It is
            #   not the case which we want to support.
            self.fail(p, schema, "Empty schema is not allowed.")
        elif "$ref" in schema:
            definition_name = schema["$ref"].replace("#/definitions/", "")
            if definition_name not in definitions:
                self.fail(p, schema,
                          "Definition '%s' is not found." % definition_name)
        else:
            self.fail(p, schema, "Wrong format.")

    @plugins.ensure_plugins_are_loaded
    def test_schema_is_valid(self):
        for p in plugin.Plugin.get_all():
            if not hasattr(p, "CONFIG_SCHEMA") or "tests.unit" in p.__module__:
                continue
            # allow only top level definitions
            definitions = p.CONFIG_SCHEMA.get("definitions", {})
            for definition in definitions.values():
                self._check_item(p, definition, definitions)

            # check schema itself
            self._check_item(p, p.CONFIG_SCHEMA, definitions)



# Copyright (c) 2016 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""Init migration

Revision ID: ca3626f62937
Revises:
Create Date: 2016-01-07 00:27:39.687814

"""
from alembic import op
import sqlalchemy as sa

from rally.common.db import api
from rally.common.db import sa_types
from rally import exceptions


# revision identifiers, used by Alembic.
revision = "ca3626f62937"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    dialect = api.get_engine().dialect

    deployments_columns = [
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("updated_at", sa.DateTime(), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("uuid", sa.String(length=36), nullable=False),
        sa.Column("parent_uuid", sa.String(length=36), nullable=True),
        sa.Column("name", sa.String(length=255), nullable=True),
        sa.Column("started_at", sa.DateTime(), nullable=True),
        sa.Column("completed_at", sa.DateTime(), nullable=True),
        sa.Column("config", sa_types.MutableJSONEncodedDict(), nullable=False),
        sa.Column("admin", sa.PickleType(), nullable=True),
        sa.Column("users", sa.PickleType(), nullable=False),
        sa.Column("enum_deployments_status", sa.Enum(
            "cleanup->failed", "cleanup->finished", "cleanup->started",
            "deploy->failed", "deploy->finished", "deploy->inconsistent",
            "deploy->init", "deploy->started", "deploy->subdeploy",
            name="enum_deploy_status"), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("name")
    ]

    if dialect.name.startswith("sqlite"):
        deployments_columns.append(
            sa.ForeignKeyConstraint(
                ["parent_uuid"], [u"deployments.uuid"],
                name="fk_parent_uuid", use_alter=True)
        )

    # commands auto generated by Alembic - please adjust!
    op.create_table("deployments", *deployments_columns)

    op.create_index("deployment_parent_uuid", "deployments",
                    ["parent_uuid"], unique=False)

    op.create_index("deployment_uuid", "deployments", ["uuid"], unique=True)

    if not dialect.name.startswith("sqlite"):
        op.create_foreign_key("fk_parent_uuid", "deployments", "deployments",
                              ["parent_uuid"], ["uuid"])

    op.create_table(
        "workers",
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("updated_at", sa.DateTime(), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("hostname", sa.String(length=255), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("hostname", name="uniq_worker@hostname")
    )

    op.create_table(
        "resources",
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("updated_at", sa.DateTime(), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("provider_name", sa.String(length=255), nullable=True),
        sa.Column("type", sa.String(length=255), nullable=True),
        sa.Column("info", sa_types.MutableJSONEncodedDict(), nullable=False),
        sa.Column("deployment_uuid", sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(["deployment_uuid"], [u"deployments.uuid"]),
        sa.PrimaryKeyConstraint("id")
    )
    op.create_index("resource_deployment_uuid", "resources",
                    ["deployment_uuid"], unique=False)

    op.create_index("resource_provider_name", "resources",
                    ["deployment_uuid", "provider_name"], unique=False)

    op.create_index("resource_provider_name_and_type", "resources",
                    ["deployment_uuid", "provider_name", "type"],
                    unique=False)

    op.create_index("resource_type", "resources",
                    ["deployment_uuid", "type"], unique=False)

    op.create_table(
        "tasks",
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("updated_at", sa.DateTime(), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("uuid", sa.String(length=36), nullable=False),
        sa.Column("status", sa.Enum(
            "aborted", "aborting", "cleaning up", "failed", "finished",
            "init", "paused", "running", "setting up", "soft_aborting",
            "verifying", name="enum_tasks_status"), nullable=False),
        sa.Column("verification_log", sa.Text(), nullable=True),
        sa.Column("tag", sa.String(length=64), nullable=True),
        sa.Column("deployment_uuid", sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(["deployment_uuid"], [u"deployments.uuid"], ),
        sa.PrimaryKeyConstraint("id")
    )

    op.create_index("task_deployment", "tasks", ["deployment_uuid"],
                    unique=False)

    op.create_index("task_status", "tasks", ["status"], unique=False)

    op.create_index("task_uuid", "tasks", ["uuid"], unique=True)

    op.create_table(
        "verifications",
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("updated_at", sa.DateTime(), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("uuid", sa.String(length=36), nullable=False),
        sa.Column("deployment_uuid", sa.String(length=36), nullable=False),
        sa.Column("status", sa.Enum(
            "aborted", "aborting", "cleaning up", "failed", "finished",
            "init", "paused", "running", "setting up", "soft_aborting",
            "verifying", name="enum_tasks_status"), nullable=False),
        sa.Column("set_name", sa.String(length=20), nullable=True),
        sa.Column("tests", sa.Integer(), nullable=True),
        sa.Column("errors", sa.Integer(), nullable=True),
        sa.Column("failures", sa.Integer(), nullable=True),
        sa.Column("time", sa.Float(), nullable=True),
        sa.ForeignKeyConstraint(["deployment_uuid"], [u"deployments.uuid"], ),
        sa.PrimaryKeyConstraint("id")
    )

    op.create_index("verification_uuid", "verifications", ["uuid"],
                    unique=True)

    op.create_table(
        "task_results",
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("updated_at", sa.DateTime(), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("key", sa_types.MutableJSONEncodedDict(), nullable=False),
        sa.Column("data", sa_types.MutableJSONEncodedDict(), nullable=False),
        sa.Column("task_uuid", sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(["task_uuid"], ["tasks.uuid"], ),
        sa.PrimaryKeyConstraint("id")
    )

    op.create_table(
        "verification_results",
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("updated_at", sa.DateTime(), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("verification_uuid", sa.String(length=36), nullable=True),
        sa.Column("data", sa_types.MutableJSONEncodedDict(), nullable=False),
        sa.ForeignKeyConstraint(["verification_uuid"], ["verifications.uuid"]),
        sa.PrimaryKeyConstraint("id")
    )
    # end Alembic commands


def downgrade():
    raise exceptions.DowngradeNotSupported()



# Copyright 2013: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import collections
import copy
import multiprocessing
import time

from rally.common import logging
from rally.common.plugin import plugin
from rally.common import utils as rutils
from rally.common import validation
from rally.task import scenario
from rally.task import types
from rally.task import utils


LOG = logging.getLogger(__name__)
configure = plugin.configure


def format_result_on_timeout(exc, timeout):
    return {
        "duration": timeout,
        "idle_duration": 0,
        "output": {"additive": [], "complete": []},
        "atomic_actions": [],
        "error": utils.format_exc(exc)
    }


def _get_scenario_context(iteration, context_obj):
    context_obj = copy.deepcopy(context_obj)
    context_obj["iteration"] = iteration + 1  # Numeration starts from `1'
    return context_obj


def _run_scenario_once(cls, method_name, context_obj, scenario_kwargs,
                       event_queue):
    iteration = context_obj["iteration"]
    event_queue.put({
        "type": "iteration",
        "value": iteration,
    })

    # provide arguments isolation between iterations
    scenario_kwargs = copy.deepcopy(scenario_kwargs)

    LOG.info("Task %(task)s | ITER: %(iteration)s START" %
             {"task": context_obj["task"]["uuid"], "iteration": iteration})

    scenario_inst = cls(context_obj)
    error = []
    try:
        with rutils.Timer() as timer:
            getattr(scenario_inst, method_name)(**scenario_kwargs)
    except Exception as e:
        error = utils.format_exc(e)
        if logging.is_debug():
            LOG.exception("Iteration %s raised Exception" % iteration)
    finally:
        status = "Error %s: %s" % tuple(error[0:2]) if error else "OK"
        LOG.info("Task %(task)s | ITER: %(iteration)s END: %(status)s" %
                 {"task": context_obj["task"]["uuid"], "iteration": iteration,
                  "status": status})

        return {"duration": timer.duration() - scenario_inst.idle_duration(),
                "timestamp": timer.timestamp(),
                "idle_duration": scenario_inst.idle_duration(),
                "error": error,
                "output": scenario_inst._output,
                "atomic_actions": scenario_inst.atomic_actions()}


def _worker_thread(queue, cls, method_name, context_obj, scenario_kwargs,
                   event_queue):
    queue.put(_run_scenario_once(cls, method_name, context_obj,
                                 scenario_kwargs, event_queue))


def _log_worker_info(**info):
    """Log worker parameters for debugging.

    :param info: key-value pairs to be logged
    """
    info_message = "\n\t".join(["%s: %s" % (k, v)
                                for k, v in info.items()])
    LOG.debug("Starting a worker.\n\t%s" % info_message)


@validation.add_default("jsonschema")
@plugin.base()
class ScenarioRunner(plugin.Plugin, validation.ValidatablePluginMixin,
                     metaclass=abc.ABCMeta):
    """Base class for all scenario runners.

    Scenario runner is an entity that implements a certain strategy of
    launching scenarios plugins, e.g. running them continuously or
    periodically for a given number of times or seconds.
    These strategies should be implemented in subclasses of ScenarioRunner
    in the_run_scenario() method.
    """

    CONFIG_SCHEMA = {
        "type": "object",
        "properties": {
            "type": {"type": "string"},
        },
        "required": ["type"],
        "additionalProperties": True
    }

    def __init__(self, task, config, batch_size=0):
        """Runner constructor.

        It sets task and config to local variables. Also initialize
        result_queue, where results will be put by _send_result method.

        :param task: Instance of objects.Task
        :param config: Dict with runner section of input task
        """
        self.task = task
        self.config = config
        self.result_queue = collections.deque()
        self.event_queue = collections.deque()
        self.aborted = multiprocessing.Event()
        self.run_duration = 0
        self.batch_size = batch_size
        self.result_batch = []

    @abc.abstractmethod
    def _run_scenario(self, cls, method_name, context, args):
        """Runs the specified scenario with given arguments.

        :param cls: The Scenario class where the scenario is implemented
        :param method_name: Name of the method that implements the scenario
        :param context: dict object that contains data created
                        by contexts plugins
        :param args: Arguments to call the scenario plugin with

        :returns: List of results fore each single scenario iteration,
                  where each result is a dictionary
        """

    def run(self, name, context, args):
        scenario_plugin = scenario.Scenario.get(name)

        # NOTE(boris-42): processing @types decorators
        args = types.preprocess(name, context, args)

        with rutils.Timer() as timer:
            # TODO(boris-42): remove method_name argument, now it's always run
            self._run_scenario(scenario_plugin, "run", context, args)

        self.run_duration = timer.duration()

    def abort(self):
        """Abort the execution of further scenario iterations."""
        self.aborted.set()

    @staticmethod
    def _create_process_pool(processes_to_start, worker_process,
                             worker_args_gen):
        """Create a pool of processes with some defined target function.

        :param processes_to_start: number of processes to create in the pool
        :param worker_process: target function for all processes in the pool
        :param worker_args_gen: generator of arguments for the target function
        :returns: the process pool as a deque
        """
        process_pool = collections.deque()

        for i in range(processes_to_start):
            kwrgs = {"processes_to_start": processes_to_start,
                     "processes_counter": i}
            process = multiprocessing.Process(target=worker_process,
                                              args=next(worker_args_gen),
                                              kwargs={"info": kwrgs})
            process.start()
            process_pool.append(process)

        return process_pool

    def _join_processes(self, process_pool, result_queue, event_queue):
        """Join the processes in the pool and send their results to the queue.

        :param process_pool: pool of processes to join
        :param result_queue: multiprocessing.Queue that receives the results
        :param event_queue: multiprocessing.Queue that receives the events
        """
        while process_pool:
            while process_pool and not process_pool[0].is_alive():
                process_pool.popleft().join()

            if result_queue.empty() and event_queue.empty():
                # sleep a bit to avoid 100% usage of CPU by this method
                time.sleep(0.01)

            while not event_queue.empty():
                self.send_event(**event_queue.get())

            while not result_queue.empty():
                self._send_result(result_queue.get())

        self._flush_results()
        result_queue.close()
        event_queue.close()

    def _flush_results(self):
        if self.result_batch:
            sorted_batch = sorted(self.result_batch)
            self.result_queue.append(sorted_batch)
            del self.result_batch[:]

    def _send_result(self, result):
        """Store partial result to send it to consumer later.

        :param result: Result dict to be sent. It should match the
                       ScenarioRunnerResult schema, otherwise
                       ValidationError is raised.
        """

        if not self.task.result_has_valid_schema(result):
            LOG.warning(
                "Task %(task)s | Runner `%(runner)s` is trying to send "
                "results in wrong format"
                % {"task": self.task["uuid"], "runner": self.get_name()})
            return

        self.result_batch.append(result)

        if len(self.result_batch) >= self.batch_size:
            sorted_batch = sorted(self.result_batch,
                                  key=lambda r: result["timestamp"])
            self.result_queue.append(sorted_batch)
            del self.result_batch[:]

    def send_event(self, type, value=None):
        """Store event to send it to consumer later.

        :param type: Event type
        :param value: Optional event data
        """
        self.event_queue.append({"type": type,
                                 "value": value})

    def _log_debug_info(self, **info):
        """Log runner parameters for debugging.

        The method logs the runner name, the task id as well as the values
        passed as arguments.

        :param info: key-value pairs to be logged
        """
        info_message = "\n\t".join(["%s: %s" % (k, v)
                                    for k, v in info.items()])
        LOG.debug("Starting the %(name)s runner (task UUID: %(task)s)."
                  "\n\t%(info)s"
                  % {"name": self._meta_get("name"),
                     "task": self.task["uuid"],
                     "info": info_message})



# Copyright 2016: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import collections
import threading

from rally.common import logging
from rally.common.plugin import plugin
from rally.common import utils as rutils
from rally.common import validation
from rally import consts
from rally import exceptions
from rally.task.processing import charts
from rally.task import utils


LOG = logging.getLogger(__name__)


configure = plugin.configure


class HookExecutor(object):
    """Runs hooks and collects results from them."""

    def __init__(self, config, task):
        self.config = config
        self.task = task

        self.triggers = collections.defaultdict(list)
        for hook_cfg in config.get("hooks", []):
            action_name = hook_cfg["action"][0]
            trigger_name = hook_cfg["trigger"][0]
            action_cls = HookAction.get(action_name)
            trigger_obj = HookTrigger.get(
                trigger_name)(hook_cfg, self.task, action_cls)
            event_type = trigger_obj.get_listening_event()
            self.triggers[event_type].append(trigger_obj)

        if "time" in self.triggers:
            self._timer_thread = threading.Thread(target=self._timer_method)
            self._timer_stop_event = threading.Event()

    def _timer_method(self):
        """Timer thread method.

        It generates events with type "time" to inform HookExecutor
        about how many time passed since beginning of the first iteration.
        """
        stopwatch = rutils.Stopwatch(stop_event=self._timer_stop_event)
        stopwatch.start()
        seconds_since_start = 0
        while not self._timer_stop_event.isSet():
            self.on_event(event_type="time", value=seconds_since_start)
            seconds_since_start += 1
            stopwatch.sleep(seconds_since_start)

    def _start_timer(self):
        self._timer_thread.start()

    def _stop_timer(self):
        self._timer_stop_event.set()
        if self._timer_thread.ident is not None:
            self._timer_thread.join()

    def on_event(self, event_type, value):
        """Notify about event.

        This method should be called to inform HookExecutor that
        particular event occurred.
        It runs hooks configured for event.
        """
        if "time" in self.triggers:
            # start timer on first iteration
            if event_type == "iteration" and value == 1:
                self._start_timer()

        for trigger_obj in self.triggers[event_type]:
            started = trigger_obj.on_event(event_type, value)
            if started:
                LOG.info("Hook %s is trigged for Task %s by %s=%s"
                         % (trigger_obj.hook_cls.__name__, self.task["uuid"],
                            event_type, value))

    def results(self):
        """Returns list of dicts with hook results."""
        if "time" in self.triggers:
            self._stop_timer()
        results = []
        for triggers_group in self.triggers.values():
            for trigger_obj in triggers_group:
                results.append(trigger_obj.get_results())
        return results


@validation.add_default("jsonschema")
@plugin.base()
class HookAction(plugin.Plugin, validation.ValidatablePluginMixin,
                 metaclass=abc.ABCMeta):
    """Factory for hook classes."""

    CONFIG_SCHEMA = {"type": "null"}

    def __init__(self, task, config, triggered_by):
        self.task = task
        self.config = config
        self._triggered_by = triggered_by
        self._thread = threading.Thread(target=self._thread_method)
        self._started_at = 0.0
        self._finished_at = 0.0
        self._result = {
            "status": consts.HookStatus.SUCCESS,
            "started_at": self._started_at,
            "finished_at": self._finished_at,
            "triggered_by": self._triggered_by,
        }

    def _thread_method(self):
        # Run hook synchronously
        self.run_sync()

    def set_error(self, exception_name, description, details):
        """Set error related information to result.

        :param exception_name: name of exception as string
        :param description: short description as string
        :param details: any details as string
        """
        self.set_status(consts.HookStatus.FAILED)
        self._result["error"] = {"etype": exception_name,
                                 "msg": description, "details": details}

    def set_status(self, status):
        """Set status to result."""
        self._result["status"] = status

    def add_output(self, additive=None, complete=None):
        """Save custom output.

        :param additive: dict with additive output
        :param complete: dict with complete output
        :raises RallyException: if output has wrong format
        """
        if "output" not in self._result:
            self._result["output"] = {"additive": [], "complete": []}
        for key, value in (("additive", additive), ("complete", complete)):
            if value:
                message = charts.validate_output(key, value)
                if message:
                    raise exceptions.RallyException(message)
                self._result["output"][key].append(value)

    def run_async(self):
        """Run hook asynchronously."""
        self._thread.start()

    def run_sync(self):
        """Run hook synchronously."""
        try:
            with rutils.Timer() as timer:
                self.run()
        except Exception as exc:
            LOG.exception("Hook %s failed during run." % self.get_name())
            self.set_error(*utils.format_exc(exc))

        self._started_at = timer.timestamp()
        self._result["started_at"] = self._started_at
        self._finished_at = timer.finish_timestamp()
        self._result["finished_at"] = self._finished_at

    @abc.abstractmethod
    def run(self):
        """Run method.

        This method should be implemented in plugin.

        Hook plugin should call following methods to set result:
            set_status - to set hook execution status
        Optionally the following methods should be called:
            set_error - to indicate that there was an error;
                        automatically sets hook execution status to 'failed'
            add_output - provide data for report
        """

    def result(self):
        """Wait and return result of hook."""
        if self._thread.ident is not None:
            # hook is still running, wait for result
            self._thread.join()
        return self._result


@validation.add_default("jsonschema")
@plugin.base()
class HookTrigger(plugin.Plugin, validation.ValidatablePluginMixin,
                  metaclass=abc.ABCMeta):
    """Factory for hook trigger classes."""

    CONFIG_SCHEMA = {"type": "null"}

    def __init__(self, hook_cfg, task, hook_cls):
        self.hook_cfg = hook_cfg
        self.config = self.hook_cfg["trigger"][1]
        self.task = task
        self.hook_cls = hook_cls
        self._runs = []

    @abc.abstractmethod
    def get_listening_event(self):
        """Returns event type to listen."""

    def on_event(self, event_type, value=None):
        """Launch hook on specified event."""
        LOG.info("Hook action %s is triggered for Task %s by %s=%s"
                 % (self.hook_cls.get_name(), self.task["uuid"],
                    event_type, value))
        action_cfg = self.hook_cfg["action"][1]
        action = self.hook_cls(self.task, action_cfg,
                               {"event_type": event_type, "value": value})
        action.run_async()
        self._runs.append(action)

    def get_results(self):
        results = {"config": self.hook_cfg,
                   "results": [],
                   "summary": {}}
        for action in self._runs:
            action_result = action.result()
            results["results"].append(action_result)
            results["summary"].setdefault(action_result["status"], 0)
            results["summary"][action_result["status"]] += 1
        return results



# Copyright 2014: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import threading

from rally.common import logging


LOG = logging.getLogger(__name__)


def _consumer(consume, queue):
    """Infinity worker that consumes tasks from queue.

    :param consume: method that consumes an object removed from the queue
    :param queue: deque object to popleft() objects from
    """
    cache = {}
    while True:
        if not queue:
            break
        else:
            try:
                args = queue.popleft()
            except IndexError:
                # consumed by other thread
                continue
        try:
            consume(cache, args)
        except Exception as e:
            msg = "Failed to consume a task from the queue"
            if logging.is_debug():
                LOG.exception(msg)
            else:
                LOG.warning("%s: %s" % (msg, e))


def _publisher(publish, queue):
    """Calls a publish method that fills queue with jobs.

    :param publish: method that fills the queue
    :param queue: deque object to be filled by the publish() method
    """
    try:
        publish(queue)
    except Exception as e:
        msg = "Failed to publish a task to the queue"
        if logging.is_debug():
            LOG.exception(msg)
        else:
            LOG.warning("%s: %s" % (msg, e))


def run(publish, consume, consumers_count=1):
    """Run broker.

    publish() put to queue, consume() process one element from queue.

    When publish() is finished and elements from queue are processed process
    is finished all consumers threads are cleaned.

    :param publish: Function that puts values to the queue
    :param consume: Function that processes a single value from the queue
    :param consumers_count: Number of consumers
    """
    queue = collections.deque()
    _publisher(publish, queue)

    consumers = []
    for i in range(consumers_count):
        consumer = threading.Thread(target=_consumer, args=(consume, queue))
        consumer.start()
        consumers.append(consumer)

    for consumer in consumers:
        consumer.join()



# Copyright 2018:  ZTE Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import subprocess

import testtools

from rally.utils import encodeutils


class CLITestCase(testtools.TestCase):

    def test_rally_cli(self):
        try:
            subprocess.check_output(["rally"], stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            output = encodeutils.safe_decode(e.output)
        else:
            self.fail("It should ve non-zero exit code.")

        self.assertIn("the following arguments are required: category", output)

    def test_version_cli(self):
        output = encodeutils.safe_decode(
            subprocess.check_output(["rally", "version"],
                                    stderr=subprocess.STDOUT))
        self.assertIn("Rally version:", output)



# Copyright 2013: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fixtures
from fixtures._fixtures.tempdir import TempDir
import os
import uuid

import mock
import testtools

from rally.common import cfg
from rally.common import db
from rally import plugins
from tests.unit import fakes


class TempHomeDir(TempDir):
    """Create a temporary directory and set it as $HOME

    :ivar path: the path of the temporary directory.
    """

    def _setUp(self):
        super(TempHomeDir, self)._setUp()
        self.useFixture(fixtures.EnvironmentVariable("HOME", self.path))


class DatabaseFixture(cfg.fixture.Config):
    """Create clean DB before starting test."""
    def setUp(self):
        super(DatabaseFixture, self).setUp()
        db_url = os.environ.get("RALLY_UNITTEST_DB_URL", "sqlite://")
        db.engine_reset()
        self.conf.set_default("connection", db_url, group="database")
        db.schema.schema_cleanup()
        db.schema.schema_create()


class TestCase(testtools.TestCase):
    """Test case base class for all unit tests."""

    def __init__(self, *args, **kwargs):
        super(TestCase, self).__init__(*args, **kwargs)

        # This is the number of characters shown when two objects do not
        # match for assertDictEqual, assertMultiLineEqual, and
        # assertSequenceEqual. The default is 640 which is too
        # low for comparing most dicts
        self.maxDiff = 10000

    def setUp(self):
        super(TestCase, self).setUp()
        self.addCleanup(mock.patch.stopall)
        plugins.load()
        self.useFixture(TempHomeDir())

    def _test_atomic_action_timer(self, atomic_actions, name, count=1,
                                  parent=[]):

        if parent:
            is_found = False
            for action in atomic_actions:
                if action["name"] == parent[0]:
                    is_found = True
                    self._test_atomic_action_timer(action["children"],
                                                   name, count=count,
                                                   parent=parent[1:])
            if not is_found:
                self.fail("The parent action %s can not be found."
                          % parent[0])
        else:
            actual_count = 0
            for atomic_action in atomic_actions:
                if atomic_action["name"] == name:
                    self.assertIsInstance(atomic_action["started_at"], float)
                    self.assertIsInstance(atomic_action["finished_at"], float)
                    actual_count += 1
            if count != actual_count:
                self.fail("%(count)d count is expected for atomic action"
                          " %(name)s, the actual count"
                          " is %(actual_count)d."
                          % {"name": name, "count": count,
                             "actual_count": actual_count})

    def assertSequenceEqual(self, iterable_1, iterable_2, msg=None):
        self.assertEqual(tuple(iterable_1), tuple(iterable_2), msg)

    _IS_EMPTY_MSG = "Iterable is not empty"

    def assertIsEmpty(self, iterable, msg=None):
        if len(iterable):
            if msg:
                msg = "%s : %s" % (self._IS_EMPTY_MSG, msg)
            else:
                msg = self._IS_EMPTY_MSG
            raise self.failureException(msg)


class DBTestCase(TestCase):
    """Base class for tests which use DB."""

    def setUp(self):
        super(DBTestCase, self).setUp()
        self.useFixture(DatabaseFixture())


# TODO(boris-42): This should be moved to test.plugins.test module
#                 or similar

class ScenarioTestCase(TestCase):
    """Base class for Scenario tests using mocked self.clients."""
    task_utils = "rally.task.utils"
    patch_task_utils = True

    def client_factory(self, client_type, version=None, admin=False):
        """Create a new client object."""
        return mock.MagicMock(client_type=client_type,
                              version=version,
                              admin=admin)

    def clients(self, client_type, version=None, admin=False):
        """Get a mocked client."""
        key = (client_type, version, admin)
        if key not in self._clients:
            self._clients[key] = self.client_factory(client_type,
                                                     version=version,
                                                     admin=admin)
        return self._clients[key]

    def admin_clients(self, client_type, version=None):
        """Get a mocked admin client."""
        return self.clients(client_type, version=version, admin=True)

    def client_created(self, client_type, version=None, admin=False):
        """Determine if a client has been created.

        This can be used to see if a scenario calls
        'self.clients("foo")', without checking to see what was done
        with the client object returned by that call.
        """
        key = (client_type, version, admin)
        return key in self._clients

    def get_client_mocks(self):
        base_path = "rally.plugins.openstack"

        return [
            mock.patch(
                "%s.scenario.OpenStackScenario.clients" % base_path,
                mock.Mock(side_effect=self.clients)),
            mock.patch(
                "%s.scenario.OpenStackScenario.admin_clients" % base_path,
                mock.Mock(side_effect=self.admin_clients))
        ]

    def get_test_context(self):
        return get_test_context()

    def setUp(self):
        super(ScenarioTestCase, self).setUp()
        if self.patch_task_utils:
            self.mock_resource_is = fixtures.MockPatch(
                self.task_utils + ".resource_is")
            self.mock_get_from_manager = fixtures.MockPatch(
                self.task_utils + ".get_from_manager")
            self.mock_wait_for = fixtures.MockPatch(
                self.task_utils + ".wait_for")
            self.mock_wait_for_delete = fixtures.MockPatch(
                self.task_utils + ".wait_for_delete")
            self.mock_wait_for_status = fixtures.MockPatch(
                self.task_utils + ".wait_for_status")
            self.useFixture(self.mock_resource_is)
            self.useFixture(self.mock_get_from_manager)
            self.useFixture(self.mock_wait_for)
            self.useFixture(self.mock_wait_for_delete)
            self.useFixture(self.mock_wait_for_status)

        self.mock_sleep = fixtures.MockPatch("time.sleep")
        self.useFixture(self.mock_sleep)

        self._clients = {}
        self._client_mocks = self.get_client_mocks()

        for patcher in self._client_mocks:
            patcher.start()

        self.context = self.get_test_context()

    def tearDown(self):
        for patcher in self._client_mocks:
            patcher.stop()
        super(ScenarioTestCase, self).tearDown()


class ContextClientAdapter(object):
    def __init__(self, endpoint, test_case):
        self.endpoint = endpoint
        self.test_case = test_case

    def mock_client(self, name, version=None):
        admin = self.endpoint.startswith("admin")
        client = self.test_case.clients(name, version=version, admin=admin)
        if not isinstance(client.return_value, mock.Mock):
            return client.return_value
        if client.side_effect is not None:
            # NOTE(pboldin): if a client has side_effects that means the
            # user wants some of the returned values overrided (look at
            # the test_existing_users for instance)
            return client()
        return client

    def __getattr__(self, name):
        # NOTE(pboldin): __getattr__ magic is called last, after the value
        # were looked up for in __dict__
        return lambda version=None: self.mock_client(name, version)


class ContextTestCase(ScenarioTestCase):
    def setUp(self):
        super(ContextTestCase, self).setUp()

        self._adapters = {}

    def context_client(self, endpoint, api_info=None):
        if endpoint not in self._adapters:
            self._adapters[endpoint] = ContextClientAdapter(endpoint, self)
        return self._adapters[endpoint]

    def get_client_mocks(self):
        return [
            mock.patch(
                "rally.plugins.openstack.osclients.Clients",
                mock.Mock(side_effect=self.context_client))
        ]


class FakeClientsScenarioTestCase(ScenarioTestCase):
    """Base class for Scenario tests using fake (not mocked) self.clients."""

    def client_factory(self, client_type, version=None, admin=False):
        return getattr(self._fake_clients, client_type)()

    def setUp(self):
        super(FakeClientsScenarioTestCase, self).setUp()
        self._fake_clients = fakes.FakeClients()


def get_test_context(**kwargs):
    kwargs["task"] = {"uuid": str(uuid.uuid4())}
    kwargs["owner_id"] = str(uuid.uuid4())
    return kwargs



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from rally.env import platform


@platform.configure(name="good", platform="fake")
class GoodPlatform(platform.Platform):

    CONFIG_SCHEMA = {}

    def create(self):
        return {}, {}

    def destroy(self):
        pass

    def cleanup(self, task_uuid=None):
        return {
            "message": "Coming soon!",
            "discovered": 0,
            "deleted": 0,
            "failed": 0,
            "resources": {},
            "errors": []
        }

    def check_health(self):
        return {"available": True}

    def info(self):
        return {"info": {"a": 1}}

    @classmethod
    def create_spec_from_sys_environ(cls, sys_environ):

        spec = {
            "auth_url": sys_environ.get("OS_AUTH_URL"),
            "username": sys_environ.get("OS_USERNAME"),
            "password": sys_environ.get("OS_PASSWORD")
        }
        return {"spec": spec, "available": True, "message": "Available"}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ast
import os

from tests.unit import test


class DDTDecoratorChecker(ast.NodeVisitor):
    """Visit an AST tree looking for classes lacking the ddt.ddt decorator.

    DDT uses decorators on test case functions to supply different
    test data, but if the class that those functions are members of is
    not decorated with @ddt.ddt, then the data expansion never happens
    and the tests are incomplete. This is very easy to miss both when
    writing and when reviewing code, so this visitor ensures that
    every class that contains a function decorated with a @ddt.*
    decorator is itself decorated with @ddt.ddt
    """
    def __init__(self):
        self.classes = []
        self.errors = {}

    @classmethod
    def _get_name(cls, node):
        if isinstance(node, ast.Name):
            return node.id
        if isinstance(node, ast.Attribute):
            return cls._get_name(node.value) + "." + node.attr
        return ""

    def _is_ddt(self, cls_node):
        return "ddt.ddt" in (self._get_name(d)
                             for d in cls_node.decorator_list)

    def visit_ClassDef(self, node):
        self.classes.append(node)
        self.generic_visit(node)
        self.classes.pop()

    def visit_FunctionDef(self, node):
        if not self.classes:
            # NOTE(stpierre): we only care about functions that are
            # defined inside of classes
            return
        cls = self.classes[-1]
        if cls.name in self.errors:
            # NOTE(stpierre): if this class already has been found to
            # be in error, ignore the rest of its functions
            return
        for decorator in node.decorator_list:
            if not isinstance(decorator, ast.Call):
                continue
            funcname = self._get_name(decorator.func)
            if funcname.startswith("ddt."):
                if not self._is_ddt(cls):
                    msg = ("Class %s has functions that use DDT, "
                           "but is not decorated with `ddt.ddt`" %
                           cls.name)
                    self.errors[cls.name] = {
                        "lineno": node.lineno,
                        "message": msg
                    }


class DDTDecoratorCheckerTestCase(test.TestCase):
    tests_path = os.path.join(os.path.dirname(__file__))

    def test_ddt_class_decorator(self):
        """Classes with DDT-decorated functions have ddt.ddt class decorator.

        """
        errors = []

        for dirname, dirnames, filenames in os.walk(self.tests_path):
            for filename in filenames:
                if not (filename.startswith("test_")
                        and filename.endswith(".py")):
                    continue

                filename = os.path.relpath(os.path.join(dirname, filename))

                with open(filename, "rb") as fh:
                    try:
                        tree = ast.parse(fh.read(), filename)
                    except TypeError as err:
                        errors.append({"message": str(err),
                                       "filename": filename,
                                       "lineno": -1})

                visitor = DDTDecoratorChecker()
                visitor.visit(tree)
                errors.extend(
                    dict(filename=filename, **error)
                    for error in visitor.errors.values())

        if errors:
            msg = [""]
            for error in errors:
                msg.extend([
                    "Errors at %(filename)s line %(lineno)d: %(message)s" % {
                        "message": error["message"],
                        "filename": error["filename"],
                        "lineno": error["lineno"]},
                    ""])
            self.fail("\n".join(msg))



# Copyright 2014: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import collections

from rally.common import cfg
from rally.common import logging
from rally.common.plugin import plugin
from rally.common import utils
from rally.common import validation
from rally.task import atomic
from rally.task import functional
from rally.task import utils as task_utils


LOG = logging.getLogger(__name__)
CONF = cfg.CONF
CONF_OPTS = [
    cfg.StrOpt(
        "context_resource_name_format",
        help="Template is used to generate random names of resources. X is "
             "replaced with random latter, amount of X can be adjusted")
]
CONF.register_opts(CONF_OPTS)


def configure(name, order, platform="default", hidden=False):
    """Context class wrapper.

    Each context class has to be wrapped by configure() wrapper. It
    sets essential configuration of context classes. Actually this wrapper just
    adds attributes to the class.

    :param name: Name of the class, used in the input task
    :param platform: str plugin's platform
    :param order: As far as we can use multiple context classes that sometimes
                  depend on each other we have to specify order of execution.
                  Contexts with smaller order are run first
    :param hidden: If it is true you won't be able to specify context via
                   task config
    """

    def wrapper(cls):
        cls = plugin.configure(name=name, platform=platform,
                               hidden=hidden)(cls)
        cls._meta_set("order", order)
        return cls

    return wrapper


def add_default_context(name, config):
    """Add default context that is inherit by all children plugins.

    :param name: str, name of the validator plugin
    :param kwargs: dict, arguments used to initialize validator class
        instance
    """

    def wrapper(plugin):
        plugin._default_meta_setdefault("default_context", {})
        plugin._default_meta_get("default_context")[name] = config
        return plugin

    return wrapper


# TODO(andreykurilin): BaseContext is used by Task and Verification and should
#                      be moved to common place
class BaseContext(plugin.Plugin, functional.FunctionalMixin,
                  utils.RandomNameGeneratorMixin, atomic.ActionTimerMixin,
                  metaclass=abc.ABCMeta):
    """This class is a factory for context classes.

    Every context class should be a subclass of this class and implement
    2 abstract methods: setup() and cleanup()

    It covers:
        1) proper setting up of context config
        2) Auto discovering & get by name
        3) Validation by CONFIG_SCHEMA
        4) Order of context creation

    """
    RESOURCE_NAME_FORMAT = "c_rally_XXXXXXXX_XXXXXXXX"

    CONFIG_SCHEMA = {"type": "null"}

    def __init__(self, ctx):
        super(BaseContext, self).__init__()
        config = ctx.get("config", {})
        if self.get_name() in config:
            # TODO(boris-42): Fix tests, code is always using fullnames
            config = config[self.get_name()]
        else:
            # TODO(boris-42): use [] instead of get() context full name is
            #                 always presented.
            config = config.get(self.get_fullname(), {})
        # NOTE(amaretskiy): self.config is a constant data and must be
        #                   immutable or write-protected type to prevent
        #                   unexpected changes in runtime
        if isinstance(config, dict):
            if hasattr(self, "DEFAULT_CONFIG"):
                for key, value in self.DEFAULT_CONFIG.items():
                    config.setdefault(key, value)
            self.config = utils.LockedDict(config)
        elif isinstance(config, list):
            self.config = tuple(config)
        else:
            # NOTE(amaretskiy): It is improbable that config can be a None,
            #                   number, boolean or even string,
            #                   however we handle this
            self.config = config
        self.context = ctx
        self.env = self.context.get("env", {})

    @classmethod
    def get_order(cls):
        return cls._meta_get("order")

    @abc.abstractmethod
    def setup(self):
        """Prepare environment for test.

        This method is executed only once before load generation.

        self.config contains input arguments of this context
        self.context contains information that will be passed to scenario

        The goal of this method is to perform all operation to prepare
        environment and store information to self.context that is required
        by scenario.
        """

    @abc.abstractmethod
    def cleanup(self):
        """Clean up environment after load generation.

        This method is run once after load generation is done to cleanup
        environment.

        self.config contains input arguments of this context
        self.context contains information that was passed to scenario
        """

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.cleanup()

    def __eq__(self, other):
        return self.get_order() == other.get_order()

    def __lt__(self, other):
        return self.get_order() < other.get_order()

    def __gt__(self, other):
        return self.get_order() > other.get_order()

    def __le__(self, other):
        return self.get_order() <= other.get_order()

    def __ge__(self, other):
        return self.get_order() >= other.get_order()


@validation.add_default("jsonschema")
@plugin.base()
class Context(BaseContext, validation.ValidatablePluginMixin):
    """The base class for task contexts."""
    def __init__(self, ctx):
        super(Context, self).__init__(ctx)
        self.task = self.context.get("task", {})

    @classmethod
    def _get_resource_name_format(cls):
        return (CONF.context_resource_name_format
                or super(Context, cls)._get_resource_name_format())

    def get_owner_id(self):
        if "owner_id" in self.context:
            return self.context["owner_id"]
        return super(Context, self).get_owner_id()


class ContextManager(object):
    """Create context environment and run method inside it."""

    def __init__(self, context_obj):
        self._visited = []
        self.context_obj = context_obj
        self._data = collections.OrderedDict()

    def contexts_results(self):
        """Returns a list with contexts execution results."""
        return list(self._data.values())

    def _get_sorted_context_lst(self):
        ctx_lst = [Context.get(name, allow_hidden=True)
                   for name in self.context_obj["config"]]
        ctx_lst.sort(key=lambda x: x.get_order())
        return [c(self.context_obj) for c in ctx_lst]

    def _log_prefix(self):
        return "Task %s |" % self.context_obj["task"]["uuid"]

    def setup(self):
        """Creates environment by executing provided context plugins."""
        self._visited = []
        for ctx in self._get_sorted_context_lst():
            ctx_data = {
                "plugin_name": ctx.get_fullname(),
                "plugin_cfg": ctx.config,
                "setup": {
                    "started_at": None,
                    "finished_at": None,
                    "atomic_actions": None,
                    "error": None
                },
                "cleanup": {
                    "started_at": None,
                    "finished_at": None,
                    "atomic_actions": None,
                    "error": None
                }
            }
            self._data[ctx.get_fullname()] = ctx_data
            self._visited.append(ctx)
            msg = ("%(log_prefix)s Context %(name)s setup() "
                   % {"log_prefix": self._log_prefix(),
                      "name": ctx.get_fullname()})

            timer = utils.Timer()
            try:
                with timer:
                    ctx.setup()
            except Exception as exc:
                ctx_data["setup"]["error"] = task_utils.format_exc(exc)
                raise
            finally:
                ctx_data["setup"]["atomic_actions"] = ctx.atomic_actions()
                ctx_data["setup"]["started_at"] = timer.timestamp()
                ctx_data["setup"]["finished_at"] = timer.finish_timestamp()

            LOG.info("%(msg)s finished in %(duration)s"
                     % {"msg": msg, "duration": timer.duration(fmt=True)})

        return self.context_obj

    def cleanup(self):
        """Cleans up  environment by executing provided context plugins."""
        ctxlst = self._visited or self._get_sorted_context_lst()
        for ctx in ctxlst[::-1]:
            ctx.reset_atomic_actions()
            msg = ("%(log_prefix)s Context %(name)s cleanup()"
                   % {"log_prefix": self._log_prefix(),
                      "name": ctx.get_fullname()})
            # NOTE(andreykurilin): As for our code, ctx_data is
            #   always presented. The further checks for `ctx_data is None` are
            #   added just for "disaster cleanup". It is not officially
            #   presented feature and not we provide out-of-the-box, but some
            #   folks have own scripts which are based on ContextManager and
            #   it would be nice to not break them.
            ctx_data = None
            if ctx.get_fullname() in self._data:
                ctx_data = self._data[ctx.get_fullname()]

            timer = utils.Timer()
            try:
                with timer:
                    LOG.info("%s started" % msg)
                    ctx.cleanup()
                LOG.info("%(msg)s finished in %(duration)s"
                         % {"msg": msg, "duration": timer.duration(fmt=True)})
            except Exception as exc:
                LOG.exception(
                    "%(msg)s failed after %(duration)s"
                    % {"msg": msg, "duration": timer.duration(fmt=True)})
                if ctx_data is not None:
                    ctx_data["cleanup"]["error"] = task_utils.format_exc(exc)
            finally:
                if ctx_data is not None:
                    aa = ctx.atomic_actions()
                    ctx_data["cleanup"]["atomic_actions"] = aa
                    ctx_data["cleanup"]["started_at"] = timer.timestamp()
                    finished_at = timer.finish_timestamp()
                    ctx_data["cleanup"]["finished_at"] = finished_at

    def __enter__(self):
        try:
            self.setup()
        except Exception:
            self.cleanup()
            raise

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.cleanup()



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
import inspect

from rally.common.plugin import discover
from rally.common.plugin import meta
from rally import exceptions
from rally.task import atomic


def service(service_name, service_type, version, client_name=None):
    """Mark class as an implementation of partial service APIs.

    :param service_name: name of the service (e.g. Nova)
    :type service_name: str
    :param service_type: type of the service (e.g. Compute)
    :type service_type: str
    :param version: version of service (e.g. 2.1)
    :type version: str
    :param client_name: name of client for service. If None, service_name will
        be used instead.
    :type client_name: str
    """
    def wrapper(cls):
        cls._meta_init()
        cls._meta_set("name", service_name.lower())
        cls._meta_set("type", service_type.lower())
        cls._meta_set("version", str(version))
        cls._meta_set("client_name", client_name or service_name)
        return cls
    return wrapper


def compat_layer(original_impl):
    """Set class which should be unified to common interface

    :param original_impl: implementation of specific service API
    :type original_impl: cls
    """
    def wrapper(cls):
        cls._meta_init()
        cls._meta_set("impl", original_impl)
        return cls
    return wrapper


def should_be_overridden(func):
    """Mark method which should be overridden by subclasses."""
    func.require_impl = True
    return func


def method_wrapper(func):
    """Wraps service's methods with some magic

    1) Each service method should not be called with positional arguments,
       since it can lead mistakes in wrong order while writing version
       compatible code. We had such situation in KeystoneWrapper
       (see https://review.opendev.org/#/c/309470/ ):

       .. code-block:: python

           class IdentityService(Service):
               def add_role(self, role_id, user_id, project_id):
                   self._impl(role_id, user_id, project_id)

           class KeystoneServiceV2(Service):
               def add_role(self, user_id, role_id, project_id):
                   pass

           class KeystoneServiceV3(Service):
               def add_role(self, role_id, user_id, project_id):
                   pass

       Explanation of example: The signature of add_role method is
       different in KeystoneServiceV2 and KeystoneServiceV3. Since
       IdentityService uses positional arguments to make call to
       self._impl.add_role, we have swapped values of role_id and user_id in
       case of KeystoneServiceV2.

       Original code and idea are taken from `positional` library.

    2) We do not need keep atomics for some actions, for example for inner
       actions (until we start to support them). Previously, we used
       "atomic_action" argument with `if atomic_action` checks inside each
       method. To reduce number of similar if blocks, let's write them in one
       place, make the code cleaner and support such feature for all service
       methods.
    """

    @functools.wraps(func)
    def wrapper(instance, *args, **kwargs):
        args_len = len(args)

        if args_len > 1:
            message = ("%(name)s takes at most 1 positional argument "
                       "(%(given)d given)" % {"name": func.__name__,
                                              "given": args_len})

            raise TypeError(message)

        return func(instance, *args, **kwargs)

    return wrapper


class ServiceMeta(type):
    """Alternative implementation of abstract classes for Services.

    Common class of specific Service should not be hardcoded for any version of
    API. We expect that all public methods of specific common class are
    overridden in all versioned implementation.
    """
    def __new__(mcs, name, parents, dct):
        for field in dct:
            if not field.startswith("_") and callable(dct[field]):
                dct[field] = method_wrapper(dct[field])
        return super(ServiceMeta, mcs).__new__(mcs, name, parents, dct)

    def __init__(cls, name, bases, namespaces):
        super(ServiceMeta, cls).__init__(name, bases, namespaces)
        bases = [c for c in cls.__bases__ if type(c) == ServiceMeta]
        if not bases:
            # nothing to check
            return

        # obtain all properties of cls, since namespace doesn't include
        # properties of parents
        not_implemented_apis = set()
        for name, obj in inspect.getmembers(cls):
            if (getattr(obj, "require_impl", False)
                    # name in namespace means that object was introduced in cls
                    and name not in namespaces):
                # it is not overridden...
                not_implemented_apis.add(name)

        if not_implemented_apis:
            raise exceptions.RallyException(
                "%s has wrong implementation. Implementation of specific "
                "version of API should override all required methods of "
                "base service class. Missed method(s): %s." %
                (cls.__name__, ", ".join(not_implemented_apis)))


class Service(meta.MetaMixin, metaclass=ServiceMeta):
    """Base help class for Cloud Services(for example OpenStack services).

    A simple example of implementation:

    .. code-block::

        # Implementation of Keystone V2 service
        @service("keystone", service_type="identity", version="2")
        class KeystoneV2Service(Service):

            @atomic.action_timer("keystone_v2.create_tenant")
            def create_tenant(self, tenant_name):
                return self.client.tenants.create(project_name)

        # Implementation of Keystone V3 service
        @service("keystone", service_type="identity", version="3")
        class KeystoneV3Service(Service):

            @atomic.action_timer("keystone_v3.create_project")
            def create_project(self, project_name):
                return self.client.project.create(project_name)

    """

    def __init__(self, clients, name_generator=None, atomic_inst=None):
        """Initialize service class

        :param clients: instance of rally.plugins.openstack.osclients.Clients
        :param name_generator: a method for generating random names. Usually
            it is generate_random_name method of RandomNameGeneratorMixin
            instance.
        :param atomic_inst: an object to store atomic actions. Usually, it is
            `_atomic_actions` property of ActionTimerMixin instance
        """
        self._clients = clients
        self._name_generator = name_generator

        if atomic_inst is None:
            self._atomic_actions = atomic.ActionTimerMixin().atomic_actions()
        else:
            self._atomic_actions = atomic_inst

        self.version = None
        if self._meta_is_inited(raise_exc=False):
            self.version = self._meta_get("version")

    def generate_random_name(self):
        if not self._name_generator:
            raise exceptions.RallyException(
                "You cannot use `generate_random_name` method, until you "
                "initialize class with `name_generator` argument.")
        return self._name_generator()


class UnifiedService(Service):
    """Base help class for unified layer for Cloud Services

    A simple example of Identity service implementation:

    .. code-block::

        import collections


        Project = collections.namedtuple("Project", ["id", "name"])


        # Unified entry-point for Identity OpenStack service
        class Identity(UnifiedService):

            # this method is equal in UnifiedKeystoneV2 and UnifiedKeystoneV3.
            # Since there is no other implementation except Keystone, there
            # are no needs to copy-paste it.
            @classmethod
            def _is_applicable(cls, clients):
                cloud_version = clients.keystone().version.split(".")[0][1:]
            return cloud_version == impl._meta_get("version")

            def create_project(self, project_name, domain_name="Default"):
                return self._impl.create_project(project_name,
                                                 domain_name=domain_name)


        # Class which unifies raw keystone v2 data to common form
        @compat_layer(KeystoneV2Service)
        class UnifiedKeystoneV2(Identity):
            def create_project(self, project_name, domain_name="Default"):
                if domain_name.lower() != "default":
                    raise NotImplementedError(
                        "Domain functionality not implemented in Keystone v2")
                tenant = self._impl.create_tenant(project_name)
                return Project(id=tenant.id, name=tenant.name)

        # Class which unifies raw keystone v3 data to common form
        @compat_layer(KeystoneV3Service)
        class UnifiedKeystoneV3(Identity):
            def create_project(self, project_name, domain_name="Default"):
                project = self._impl.create_project(project_name,
                                                    domain_name=domain_name)
                return Project(id=project.id, name=project.name)
    """

    def __init__(self, clients, name_generator=None, atomic_inst=None):
        """Initialize service class

        :param clients: instance of rally.plugins.openstack.osclients.Clients
        :param name_generator: a method for generating random names. Usually
            it is generate_random_name method of RandomNameGeneratorMixin
            instance.
        :param atomic_inst: an object to store atomic actions. Usually, it is
            `_atomic_actions` property of ActionTimerMixin instance
        """
        super(UnifiedService, self).__init__(clients, name_generator,
                                             atomic_inst)

        if self._meta_is_inited(raise_exc=False):
            # it is an instance of compatibility layer for specific Service
            impl_cls = self._meta_get("impl")
            self._impl = impl_cls(self._clients, self._name_generator,
                                  self._atomic_actions)
            self.version = impl_cls._meta_get("version")
        else:
            # it is a base class of service
            impl_cls, _all_impls = self.discover_impl()
            if not impl_cls:
                raise exceptions.RallyException(
                    "There is no proper implementation for %s."
                    % self.__class__.__name__)
            self._impl = impl_cls(self._clients, self._name_generator,
                                  self._atomic_actions)
            self.version = self._impl.version

    def discover_impl(self):
        """Discover implementation for service

        One Service can have different implementations(not only in terms of
        versioning, for example Network service of OpenStack has Nova-network
        and Neutron implementation. they are quite different). Each of such
        implementations can support several versions. This method is designed
        to choose the proper helper class based on available services in the
        cloud and based on expected version.

        Returns a tuple with implementation class as first element, a set of
            all implementations as a second element
        """

        # find all classes with unified implementation
        impls = {cls: cls._meta_get("impl")
                 for cls in discover.itersubclasses(self.__class__)
                 if (cls._meta_is_inited(raise_exc=False)
                     and cls._meta_get("impl"))}

        service_names = {o._meta_get("name") for o in impls.values()}

        enabled_services = None
        # let's make additional calls to cloud only when we need to make a
        # decision based on available services
        if len(service_names) > 1:
            enabled_services = list(self._clients.services().values())

        for cls, impl in impls.items():
            if (enabled_services is not None
                    and impl._meta_get("name") not in enabled_services):
                continue
            if cls.is_applicable(self._clients):
                return cls, impls

        return None, impls

    @classmethod
    def is_applicable(cls, clients):
        """Check that implementation can be used in cloud."""

        if cls._meta_is_inited(raise_exc=False):
            impl = cls._meta_get("impl", cls)
            client = getattr(clients, impl._meta_get("client_name"))
            return client.choose_version() == impl._meta_get("version")
        return False


class _Resource(object):

    __slots__ = []
    _id_property = None

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __getitem__(self, item, default=None):
        return getattr(self, item, default)

    def __repr__(self):
        return "<%s id=%s>" % (self.__class__.__name__,
                               getattr(self, self._id_property, "n/a"))

    def __eq__(self, other):
        self_id = getattr(self, self._id_property)
        return (isinstance(other, self.__class__)
                and self_id == getattr(other, self._id_property))

    def _as_dict(self):
        return dict((k, self[k]) for k in self.__slots__)


def make_resource_cls(name, properties, id_property="id"):
    """Construct a resource class with limited number of properties.

    Unlike collections.namedtuple, a created class has user-friendly getitem
    method for obtaining properties.

    :param name: The name of resource (i.e image, container..)
    :param properties: The list of allowed properties
    :param id_property: The name of property which should be used as id of
        resource. By defaults, it is "id" field if such property presents in
        "properties" or first element of "properties" in other cases.
    """

    id_property = id_property if id_property in properties else properties[0]

    # NOTE(andreykurilin): call a `type` method instead of returning just raw
    #   class (create Resource class inside the method make_resource_cls and
    #   return it) allows to setup a custom name of a new class, which will be
    #   used in case of errors and etc
    return type(name.title(), (_Resource,), {"__slots__": properties,
                                             "_id_property": id_property})



# Copyright 2015: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from rally.common.plugin import plugin
from rally import exceptions
from tests.unit import test


class PluginModuleTestCase(test.TestCase):

    def test_deprecated_cls(self):

        @plugin.deprecated("God why?", "0.0.2")
        @plugin.configure(name="deprecated_class_plugin_check")
        class MyPlugin(plugin.Plugin):
            pass

        self.addCleanup(MyPlugin.unregister)

        self.assertEqual({"reason": "God why?", "rally_version": "0.0.2"},
                         MyPlugin.is_deprecated())

    def test_configure_cls(self):

        @plugin.configure(name="get_name_class_plugin")
        class MyPlugin(plugin.Plugin):
            pass

        self.addCleanup(MyPlugin.unregister)

        self.assertEqual("get_name_class_plugin", MyPlugin.get_name())

    def test_configure_two_cls_with_same_names(self):
        name = ".".join(self.id().rsplit(".", 2)[1:])

        @plugin.base()
        class FooBase(plugin.Plugin):
            pass

        @plugin.configure(name=name)
        class MyPlugin(FooBase):
            pass

        self.addCleanup(MyPlugin.unregister)

        try:
            @plugin.configure(name=name)
            class MyPlugin2(FooBase):
                pass

            self.addCleanup(MyPlugin2.unregister)

        except exceptions.PluginWithSuchNameExists:
            self.assertEqual([MyPlugin], FooBase.get_all())
        else:
            self.fail("Registration two plugins with the same names in one "
                      "platform should raise an exception.")

    def test_configure_different_bases(self):
        name = "test_configure_different_bases"

        @plugin.base()
        class OneBase(plugin.Plugin):
            pass

        @plugin.base()
        class SecondBase(plugin.Plugin):
            pass

        @plugin.configure(name, platform=name)
        class A(OneBase):
            pass

        self.addCleanup(A.unregister)

        @plugin.configure(name, platform=name)
        class B(SecondBase):
            pass

        self.addCleanup(B.unregister)

        self.assertEqual(OneBase.get(name), A)
        self.assertEqual(SecondBase.get(name), B)

    def test_get_multiple_chooses(self):
        name = "test_get_multiple_chooses"

        @plugin.base()
        class OneBase(plugin.Plugin):
            pass

        @plugin.base()
        class SecondBase(plugin.Plugin):
            pass

        @plugin.configure(name, platform=name)
        class A(OneBase):
            pass

        self.addCleanup(A.unregister)

        @plugin.configure(name, platform=name)
        class B(SecondBase):
            pass

        self.addCleanup(B.unregister)
        self.assertRaises(exceptions.MultiplePluginsFound, plugin.Plugin.get,
                          name, name)

    def test_get_multiple_chooses_but_default(self):
        @plugin.configure(self.id())
        class A(plugin.Plugin):
            pass

        self.addCleanup(A.unregister)

        @plugin.configure(self.id(), platform=self.id())
        class B(plugin.Plugin):
            pass

        self.addCleanup(B.unregister)

        self.assertEqual(A, plugin.Plugin.get(self.id()))

    def test_default_meta_for_base(self):

        @plugin.base()
        class CanUseDefaultMetaBase(plugin.Plugin):
            pass

        CanUseDefaultMetaBase._default_meta_set("a", 10)

        @plugin.configure("CantUseDefaultMetaPlugin")
        class CantUseDefaultMetaPlugin(CanUseDefaultMetaBase):
            pass

        self.addCleanup(CantUseDefaultMetaPlugin.unregister)
        self.assertRaises(ReferenceError,
                          CantUseDefaultMetaPlugin._default_meta_set, "a", 10)

        self.assertEqual(10, CantUseDefaultMetaPlugin._meta_get("a"))

    def test_default_meta_decorator(self):

        @plugin.default_meta()
        class CanUseDefaultMetaPlugin(plugin.Plugin):
            pass

        CanUseDefaultMetaPlugin._default_meta_set("a", 10)
        self.assertEqual(10, CanUseDefaultMetaPlugin._default_meta_get("a"))


@plugin.configure(name="test_base_plugin")
class BasePlugin(plugin.Plugin):
    pass


@plugin.configure(name="test_some_plugin")
class SomePlugin(BasePlugin):
    pass


@plugin.configure(name="test_my_plugin")
class MyPluginInDefault(BasePlugin):
    pass


@plugin.configure(name="test_my_plugin", platform="foo")
class MyPluginInFoo(BasePlugin):
    pass


@plugin.configure(name="test_hidden_plugin", hidden=True)
class HiddenPlugin(BasePlugin):
    pass


@plugin.deprecated("some_reason", "0.1.1")
@plugin.configure(name="test_deprecated_plugin")
class DeprecatedPlugin(BasePlugin):
    pass


class NotInitedPlugin(BasePlugin):
    pass


class PluginTestCase(test.TestCase):

    def test_unregister(self):

        @plugin.configure(name="test_some_temp_plugin")
        class SomeTempPlugin(BasePlugin):
            pass

        SomeTempPlugin.unregister()
        self.assertRaises(exceptions.PluginNotFound,
                          BasePlugin.get, "test_some_temp_plugin")

    def test_get(self):
        self.assertEqual(SomePlugin,
                         BasePlugin.get("test_some_plugin"))

    def test_get_hidden(self):
        self.assertEqual(HiddenPlugin,
                         BasePlugin.get("test_hidden_plugin",
                                        allow_hidden=True))

    def test_get_hidden_not_found(self):
        self.assertRaises(exceptions.PluginNotFound,
                          BasePlugin.get, "test_hidden_plugin")

    def test_get_not_found(self):
        self.assertRaises(exceptions.PluginNotFound,
                          BasePlugin.get, "non_existing")

    def test_get_multiple_found(self):

        @plugin.configure("test_2_plugins_with_same_name")
        class A(plugin.Plugin):
            pass

        class B(plugin.Plugin):
            pass

        self.assertRaises(exceptions.PluginWithSuchNameExists,
                          plugin.configure("test_2_plugins_with_same_name"), B)

        A.unregister()

    def test_get_multiple_found_hidden(self):

        @plugin.configure("test_2_plugins_with_same_name", hidden=True)
        class A(plugin.Plugin):
            pass

        class B(plugin.Plugin):
            pass

        self.assertRaises(exceptions.PluginWithSuchNameExists,
                          plugin.configure("test_2_plugins_with_same_name"), B)

        A.unregister()

    def test_get_name(self):
        self.assertEqual("test_some_plugin", SomePlugin.get_name())

    def test_get_all(self):
        self.assertEqual(set([SomePlugin, DeprecatedPlugin,
                              MyPluginInDefault, MyPluginInFoo]),
                         set(BasePlugin.get_all()))
        self.assertEqual([], SomePlugin.get_all())

    def test_get_all_by_name(self):
        self.assertEqual(set([MyPluginInDefault, MyPluginInFoo]),
                         set(BasePlugin.get_all(name="test_my_plugin")))

    def test_get_all_hidden(self):
        self.assertEqual(set([SomePlugin, DeprecatedPlugin, HiddenPlugin,
                              MyPluginInDefault, MyPluginInFoo]),
                         set(BasePlugin.get_all(allow_hidden=True)))

    def test_is_deprecated(self):
        self.assertFalse(SomePlugin.is_deprecated())
        self.assertEqual({"reason": "some_reason", "rally_version": "0.1.1"},
                         DeprecatedPlugin.is_deprecated())



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime as dt
import os

import mock

from rally.plugins.common.exporters import junit
from tests.unit import test


def get_tasks_results():
    task_id = "2fa4f5ff-7d23-4bb0-9b1f-8ee235f7f1c8"
    return [{
        "uuid": "task-uu-ii-dd",
        "created_at": "2017-06-04T05:14:00",
        "updated_at": "2017-06-04T05:15:15",
        "subtasks": [
            {"task_uuid": task_id,
             "workloads": [
                 {
                     "uuid": "workload-1-uuid",
                     "created_at": "2017-06-04T05:14:44",
                     "updated_at": "2017-06-04T05:15:14",
                     "task_uuid": task_id,
                     "position": 0,
                     "name": "CinderVolumes.list_volumes",
                     "full_duration": 29.969523191452026,
                     "sla_results": {"sla": []},
                     "pass_sla": True
                 },
                 {
                     "uuid": "workload-2-uuid",
                     "created_at": "2017-06-04T05:15:15",
                     "updated_at": "2017-06-04T05:16:14",
                     "task_uuid": task_id,
                     "position": 1,
                     "name": "NovaServers.list_keypairs",
                     "full_duration": 5,
                     "sla_results": {"sla": [
                         {"criterion": "Failing",
                          "success": False,
                          "detail": "ooops"},
                         {"criterion": "Ok",
                          "success": True,
                          "detail": None},
                     ]},
                     "pass_sla": False
                 },
             ]}]}]


class JUnitXMLExporterTestCase(test.TestCase):
    def setUp(self):
        super(JUnitXMLExporterTestCase, self).setUp()
        self.datetime = dt.datetime

        patcher = mock.patch("rally.plugins.common.exporters.junit.dt")
        self.dt = patcher.start()
        self.dt.datetime.strptime.side_effect = self.datetime.strptime
        self.addCleanup(patcher.stop)

    @mock.patch("rally.plugins.common.exporters.junit.version.version_string")
    def test_generate(self, mock_version_string):
        now = self.dt.datetime.utcnow.return_value
        now.strftime.return_value = "$TIME"
        mock_version_string.return_value = "$VERSION"

        with open(os.path.join(os.path.dirname(__file__),
                               "junit_report.xml")) as f:
            expected_report = f.read()

        reporter = junit.JUnitXMLExporter(get_tasks_results(),
                                          output_destination=None)
        self.assertEqual({"print": expected_report}, reporter.generate())

        reporter = junit.JUnitXMLExporter(get_tasks_results(),
                                          output_destination="path")
        self.assertEqual({"files": {"path": expected_report},
                          "open": "file://" + os.path.abspath("path")},
                         reporter.generate())



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import mock

from tests.ci import pytest_launcher
from tests.unit import test


PATH = "tests.ci.pytest_launcher"


class ExitError(Exception):
    pass


class PyTestLauncherTestCase(test.TestCase):
    def setUp(self):
        super(PyTestLauncherTestCase, self).setUp()

        sp_patcher = mock.patch("%s.subprocess" % PATH)
        self.sp = sp_patcher.start()
        self.addCleanup(sp_patcher.stop)

        exit_patcher = mock.patch("%s.exit" % PATH, side_effect=ExitError)
        self.exit = exit_patcher.start()
        self.addCleanup(exit_patcher.stop)

        os_patcher = mock.patch("%s.os" % PATH)
        self.os = os_patcher.start()
        self.addCleanup(os_patcher.stop)
        # emulate local run by default
        self.os.environ = {}
        self.os.path.join.side_effect = os.path.join
        self.os.path.abspath.side_effect = os.path.abspath
        self.os.path.expanduser.side_effect = os.path.expanduser

    def test_wrong_posargs(self):
        self.assertRaises(ExitError, pytest_launcher.main,
                          ["script name", "test_path",
                           "--posargs='posargs with spaces'"])

        self.assertFalse(self.sp.called)
        self.assertFalse(self.os.called)

    def test_parsing_path(self):
        def os_path_exists(path):
            dpath = "some/path/to/some/test"
            return dpath.startswith(path) or path == "%s/module.py" % dpath

        self.os.path.exists.side_effect = os_path_exists

        pytest_launcher.main(
            ["script_name", "some/path",
             "--posargs=some.path.to.some.test.module.TestCase.test"])

        expected_path = os.path.abspath(
            "some/path/to/some/test/module.py::TestCase::test")

        self.assertEqual(1, self.sp.check_call.call_count)
        call_args_obj = self.sp.check_call.call_args_list[0]
        call_args = call_args_obj[0]
        self.assertEqual(expected_path, call_args[0][-1])



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#
# Rally documentation build configuration file, created by
# sphinx-quickstart on Fri Jan 10 23:19:18 2014.
#
# This file is execfile() with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# auto-generated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import datetime as dt
import os
import subprocess
import sys

import rally.common.version

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.extend([
#     os.path.abspath("../.."),
#     os.path.abspath("../"),
#     os.path.abspath("./")
# ])

sys.path.insert(0, os.path.abspath("../../"))
sys.path.insert(0, os.path.abspath("../"))
sys.path.insert(0, os.path.abspath("./"))


# -- General configuration ----------------------------------------------------

on_rtd = os.environ.get("READTHEDOCS") == "True"

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = "1.0"

# Add any Sphinx extension module names here, as strings.
# They can be extensions coming with Sphinx (named "sphinx.ext.*") or your
# custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.ifconfig",
    "sphinx.ext.viewcode",
    "ext.cli_reference",
    "ext.plugin_reference",
    "ext.include_vars"
]
todo_include_todos = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = "utf-8-sig"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = u"Rally"
copyright = u"%d, OpenStack Foundation" % dt.datetime.now().year

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.


version = rally.common.version.version_string()
# The full version, including alpha/beta/rc tags.
release = rally.common.version.version_string()

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ""
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = "%B %d, %Y"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = [
    "feature_request/README.rst",
    "samples/README.rst",
    "**/README.rst"
]

# The reST default role (used for this markup: `text`) to use for all documents
# default_role = None

# If true, "()" will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []


# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
if not on_rtd:
    html_theme = "openstackrally"
else:
    html_theme = "default"
# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
if not on_rtd:
    import oslosphinx
    theme_dir = os.path.join(os.path.dirname(oslosphinx.__file__), "theme")
    html_theme_path = [theme_dir, "_templates"]
else:
    html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
if not on_rtd:
    html_static_path = ["_templates/openstackrally/_static"]
else:
    html_static_path = []

# If not "", a "Last updated on:" timestamp is inserted at every page bottom,
# using the given strftime format.
git_cmd = [
    "git", "log", "--pretty=format:'%ad, commit %h'", "--date=local", "-n1"]
html_last_updated_fmt = subprocess.check_output(git_cmd).decode("utf-8")

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
html_use_smartypants = False

# Custom sidebar templates, maps document names to template names.
html_sidebars = {"**": ["searchbox.html", "globaltoc.html"]}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ""

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = "%sdoc" % project


# -- Options for LaTeX output -------------------------------------------------

latex_elements = {
    # The paper size ("letterpaper" or "a4paper").
    # "papersize": "letterpaper",
    # The font size ("10pt", "11pt" or "12pt").
    # "pointsize": "10pt",
    # Additional stuff for the LaTeX preamble.
    # "preamble": "",
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual])
latex_documents = [
    ("index",
     "%s.tex" % project,
     u"%s Documentation" % project,
     u"OpenStack Foundation", "manual"),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output -------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
# man_pages = [
#    ("index", "rally", u"Rally Documentation",
#     [u"Rally Team"], 1)
# ]

# If true, show URL addresses after external links.
# man_show_urls = False


# -- Options for Texinfo output -----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    ("index", "Rally", u"Rally Documentation",
     u"Rally Team", "Rally",
     "Testing framework and tool for all kinds of tests",
     "Development"),
]

# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
# texinfo_domain_indices = True

# How to display URL addresses: "footnote", "no", or "inline".
# texinfo_show_urls = "footnote"



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import jsonschema
import mock

from rally.verification import reporter
from tests.unit import test


class ReporterTestCase(test.TestCase):

    def test_make(self):
        reporter_cls = mock.Mock()

        reporter_cls.return_value.generate.return_value = {}
        reporter.VerificationReporter.make(reporter_cls, None, None)

        reporter_cls.return_value.generate.return_value = {"files": {}}
        reporter.VerificationReporter.make(reporter_cls, None, None)
        reporter_cls.return_value.generate.return_value = {
            "files": {"/path/foo": "content"}}
        reporter.VerificationReporter.make(reporter_cls, None, None)

        reporter_cls.return_value.generate.return_value = {"open": "/path/foo"}
        reporter.VerificationReporter.make(reporter_cls, None, None)

        reporter_cls.return_value.generate.return_value = {"print": "foo"}
        reporter.VerificationReporter.make(reporter_cls, None, None)

        reporter_cls.return_value.generate.return_value = {
            "files": {"/path/foo": "content"}, "open": "/path/foo",
            "print": "foo"}
        reporter.VerificationReporter.make(reporter_cls, None, None)

        reporter_cls.return_value.generate.return_value = {"files": []}
        self.assertRaises(jsonschema.ValidationError,
                          reporter.VerificationReporter.make,
                          reporter_cls, None, None)

        reporter_cls.return_value.generate.return_value = {"files": ""}
        self.assertRaises(jsonschema.ValidationError,
                          reporter.VerificationReporter.make,
                          reporter_cls, None, None)

        reporter_cls.return_value.generate.return_value = {"files": {"a": {}}}
        self.assertRaises(jsonschema.ValidationError,
                          reporter.VerificationReporter.make,
                          reporter_cls, None, None)

        reporter_cls.return_value.generate.return_value = {"open": []}
        self.assertRaises(jsonschema.ValidationError,
                          reporter.VerificationReporter.make,
                          reporter_cls, None, None)

        reporter_cls.return_value.generate.return_value = {"print": []}
        self.assertRaises(jsonschema.ValidationError,
                          reporter.VerificationReporter.make,
                          reporter_cls, None, None)

        reporter_cls.return_value.generate.return_value = {"additional": ""}
        self.assertRaises(jsonschema.ValidationError,
                          reporter.VerificationReporter.make,
                          reporter_cls, None, None)



# Copyright 2013: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import io
import os

import mock

from rally.cli import cliutils
from rally.cli.commands import deployment
from rally.cli import envutils
from rally import consts
from rally import exceptions
from tests.unit import fakes
from tests.unit import test


class DeploymentCommandsTestCase(test.TestCase):
    def setUp(self):
        super(DeploymentCommandsTestCase, self).setUp()
        self.deployment = deployment.DeploymentCommands()
        self.fake_api = fakes.FakeAPI()

    @mock.patch.dict(os.environ, {"RALLY_DEPLOYMENT": "my_deployment_id"})
    @mock.patch("rally.cli.commands.deployment.DeploymentCommands.list")
    @mock.patch("rally.cli.commands.deployment.open",
                side_effect=mock.mock_open(read_data="{\"some\": \"json\"}"),
                create=True)
    def test_create(self, mock_open, mock_deployment_commands_list):
        self.deployment.create(self.fake_api, "fake_deploy", False,
                               "path_to_config.json")
        self.fake_api.deployment.create.assert_called_once_with(
            config={"some": "json"}, name="fake_deploy")

    @mock.patch.dict(os.environ, {"RALLY_DEPLOYMENT": "my_deployment_id"})
    @mock.patch("rally.cli.commands.deployment.DeploymentCommands.list")
    def test_create_empty(self, mock_deployment_commands_list):
        self.deployment.create(self.fake_api, "fake_deploy")
        self.fake_api.deployment.create.assert_called_once_with(
            config={}, name="fake_deploy")

    @mock.patch("rally.env.env_mgr.EnvManager.create_spec_from_sys_environ",
                return_value={"spec": {"auth_url": "http://fake"}})
    def test_create_fromenv(self, mock_create_spec_from_sys_environ):
        self.deployment.create(self.fake_api, "from_env", True)
        self.fake_api.deployment.create.assert_called_once_with(
            config={"auth_url": "http://fake"},
            name="from_env"
        )

    @mock.patch("rally.env.env_mgr.EnvManager.create_spec_from_sys_environ")
    def test_create_fromenv_openstack(self, mock_create_spec_from_sys_environ):

        mock_create_spec_from_sys_environ.side_effect = lambda: {
            "spec": {
                "existing@openstack": {
                    "https_key": "some key",
                    "another_key": "another"
                }
            }
        }
        mock_rally_os = mock.Mock()
        mock_rally_os.__version_tuple__ = (1, 4, 0)

        with mock.patch.dict("sys.modules",
                             {"rally_openstack": mock_rally_os}):
            self.deployment.create(self.fake_api, "from_env", True)
            self.fake_api.deployment.create.assert_called_once_with(
                config={"existing@openstack": {"another_key": "another"}},
                name="from_env"
            )

            self.fake_api.deployment.create.reset_mock()
            mock_rally_os.__version_tuple__ = (1, 5, 0)
            self.deployment.create(self.fake_api, "from_env", True)
            self.fake_api.deployment.create.assert_called_once_with(
                config={"existing@openstack": {"another_key": "another",
                                               "https_key": "some key"}},
                name="from_env"
            )

    @mock.patch("rally.cli.commands.deployment.DeploymentCommands.list")
    @mock.patch("rally.cli.commands.deployment.DeploymentCommands.use")
    @mock.patch("rally.cli.commands.deployment.open",
                side_effect=mock.mock_open(read_data="{\"uuid\": \"uuid\"}"),
                create=True)
    def test_create_and_use(self, mock_open, mock_deployment_commands_use,
                            mock_deployment_commands_list):
        self.fake_api.deployment.create.return_value = dict(uuid="uuid")
        self.deployment.create(self.fake_api, "fake_deploy", False,
                               "path_to_config.json", True)
        self.fake_api.deployment.create.assert_called_once_with(
            config={"uuid": "uuid"}, name="fake_deploy")
        mock_deployment_commands_list.assert_called_once_with(
            self.fake_api, deployment_list=[{"uuid": "uuid"}])
        mock_deployment_commands_use.assert_called_once_with(
            self.fake_api, self.fake_api.deployment.create.return_value)

    def test_recreate(self):
        deployment_id = "43924f8b-9371-4152-af9f-4cf02b4eced4"
        self.deployment.recreate(self.fake_api, deployment_id)
        self.fake_api.deployment.recreate.assert_called_once_with(
            deployment=deployment_id, config=None)

    @mock.patch("rally.cli.commands.deployment.open",
                side_effect=mock.mock_open(read_data="{\"some\": \"json\"}"),
                create=True)
    def test_recreate_config(self, mock_open):
        deployment_id = "43924f8b-9371-4152-af9f-4cf02b4eced4"
        self.deployment.recreate(self.fake_api, deployment_id,
                                 filename="my.json")
        self.fake_api.deployment.recreate.assert_called_once_with(
            deployment=deployment_id, config={"some": "json"})

    @mock.patch("rally.cli.commands.deployment.envutils.get_global")
    def test_recreate_no_deployment_id(self, mock_get_global):
        mock_get_global.side_effect = exceptions.InvalidArgumentsException
        self.assertRaises(exceptions.InvalidArgumentsException,
                          self.deployment.recreate, None)

    def test_destroy(self):
        deployment_id = "53fd0273-60ce-42e5-a759-36f1a683103e"
        self.deployment.destroy(self.fake_api, deployment_id)
        self.fake_api.deployment.destroy.assert_called_once_with(
            deployment=deployment_id)

    @mock.patch("rally.cli.commands.deployment.envutils.get_global")
    def test_destroy_no_deployment_id(self, mock_get_global):
        mock_get_global.side_effect = exceptions.InvalidArgumentsException
        self.assertRaises(exceptions.InvalidArgumentsException,
                          self.deployment.destroy, self.fake_api, None)

    @mock.patch("rally.cli.commands.deployment.cliutils.print_list")
    @mock.patch("rally.cli.commands.deployment.utils.Struct")
    @mock.patch("rally.cli.commands.deployment.envutils.get_global")
    def test_list_different_deployment_id(self, mock_get_global, mock_struct,
                                          mock_print_list):
        current_deployment_id = "26a3ce76-0efa-40e4-86e5-514574bd1ff6"
        mock_get_global.return_value = current_deployment_id
        fake_deployment_list = [
            {"uuid": "fa34aea2-ae2e-4cf7-a072-b08d67466e3e",
             "created_at": "03-12-2014",
             "name": "dep1",
             "status": "deploy->started",
             "active": "False"}]

        self.fake_api.deployment.list.return_value = fake_deployment_list
        self.deployment.list(self.fake_api)

        fake_deployment = fake_deployment_list[0]
        fake_deployment["active"] = ""
        mock_struct.assert_called_once_with(**fake_deployment)

        headers = ["uuid", "created_at", "name", "status", "active"]
        mock_print_list.assert_called_once_with([mock_struct()], headers,
                                                sortby_index=headers.index(
                                                "created_at"))

    @mock.patch("rally.cli.commands.deployment.cliutils.print_list")
    @mock.patch("rally.cli.commands.deployment.utils.Struct")
    @mock.patch("rally.cli.commands.deployment.envutils.get_global")
    def test_list_current_deployment_id(self, mock_get_global, mock_struct,
                                        mock_print_list):
        current_deployment_id = "64258e84-ffa1-4011-9e4c-aba07bdbcc6b"
        mock_get_global.return_value = current_deployment_id
        fake_deployment_list = [{"uuid": current_deployment_id,
                                 "created_at": "13-12-2014",
                                 "name": "dep2",
                                 "status": "deploy->finished",
                                 "active": "True"}]
        self.fake_api.deployment.list.return_value = fake_deployment_list
        self.deployment.list(self.fake_api)

        fake_deployment = fake_deployment_list[0]
        fake_deployment["active"] = "*"
        mock_struct.assert_called_once_with(**fake_deployment)

        headers = ["uuid", "created_at", "name", "status", "active"]
        mock_print_list.assert_called_once_with([mock_struct()], headers,
                                                sortby_index=headers.index(
                                                "created_at"))

    @mock.patch("json.dumps")
    def test_config(self, mock_json_dumps):
        deployment_id = "fa4a423e-f15d-4d83-971a-89574f892999"
        value = {"config": "config"}
        self.fake_api.deployment.get.return_value = value
        self.deployment.config(self.fake_api, deployment_id)
        mock_json_dumps.assert_called_once_with(value["config"],
                                                sort_keys=True, indent=4)
        self.fake_api.deployment.get.assert_called_once_with(
            deployment=deployment_id)

    @mock.patch("rally.cli.commands.deployment.envutils.get_global")
    def test_config_no_deployment_id(self, mock_get_global):
        mock_get_global.side_effect = exceptions.InvalidArgumentsException
        self.assertRaises(exceptions.InvalidArgumentsException,
                          self.deployment.config, self.fake_api, None)

    @mock.patch("rally.cli.commands.deployment.cliutils.print_list")
    @mock.patch("rally.cli.commands.deployment.utils.Struct")
    def test_show(self, mock_struct, mock_print_list):
        deployment_id = "b1a6153e-a314-4cb3-b63b-cf08c1a416c3"
        value = {"admin": {"auth_url": "url",
                           "username": "u",
                           "password": "p",
                           "tenant_name": "t",
                           "region_name": "r",
                           "endpoint_type": consts.EndpointType.INTERNAL},
                 "users": []}
        deployment = self.fake_api.deployment.get
        deployment.return_value = {"credentials": {"openstack": [
            {"admin": value["admin"],
             "users": []}]}}
        self.deployment.show(self.fake_api, deployment_id)
        self.fake_api.deployment.get.assert_called_once_with(
            deployment=deployment_id)

        headers = ["auth_url", "username", "password", "tenant_name",
                   "region_name", "endpoint_type"]
        fake_data = ["url", "u", "***", "t", "r", consts.EndpointType.INTERNAL]
        mock_struct.assert_called_once_with(**dict(zip(headers, fake_data)))
        mock_print_list.assert_called_once_with([mock_struct()], headers)

    @mock.patch("rally.cli.commands.deployment.envutils.get_global")
    def test_deploy_no_deployment_id(self, mock_get_global):
        mock_get_global.side_effect = exceptions.InvalidArgumentsException
        self.assertRaises(exceptions.InvalidArgumentsException,
                          self.deployment.show, None)

    @mock.patch("os.remove")
    @mock.patch("os.symlink")
    @mock.patch("os.path.exists", return_value=True)
    @mock.patch("rally.common.fileutils.update_env_file")
    def test_use(self, mock_update_env_file, mock_path_exists,
                 mock_symlink, mock_remove):
        deployment_id = "593b683c-4b16-4b2b-a56b-e162bd60f10b"
        self.fake_api.deployment.get.return_value = {
            "uuid": deployment_id,
            "credentials": {
                "openstack": [{
                    "admin": {"auth_url": "fake_auth_url",
                              "username": "fake_username",
                              "password": "fake_password",
                              "tenant_name": "fake_tenant_name",
                              "endpoint": "fake_endpoint",
                              "region_name": None}}]}}

        with mock.patch("rally.cli.commands.deployment.open", mock.mock_open(),
                        create=True) as mock_file:
            self.deployment.use(self.fake_api, deployment_id)
            self.assertEqual(3, mock_path_exists.call_count)
            mock_update_env_file.assert_has_calls([
                mock.call(os.path.expanduser("~/.rally/globals"),
                          "RALLY_DEPLOYMENT", "%s\n" % deployment_id),
                mock.call(os.path.expanduser("~/.rally/globals"),
                          "RALLY_ENV", "%s\n" % deployment_id),
            ])

            mock_file.return_value.write.assert_any_call(
                "export OS_ENDPOINT='fake_endpoint'\n")
            mock_file.return_value.write.assert_any_call(
                "export OS_AUTH_URL='fake_auth_url'\n"
                "export OS_USERNAME='fake_username'\n"
                "export OS_PASSWORD='fake_password'\n"
                "export OS_TENANT_NAME='fake_tenant_name'\n"
                "export OS_PROJECT_NAME='fake_tenant_name'\n")
            mock_symlink.assert_called_once_with(
                os.path.expanduser("~/.rally/openrc-%s" % deployment_id),
                os.path.expanduser("~/.rally/openrc"))
            mock_remove.assert_called_once_with(os.path.expanduser(
                "~/.rally/openrc"))

    @mock.patch("os.remove")
    @mock.patch("os.symlink")
    @mock.patch("os.path.exists", return_value=True)
    @mock.patch("rally.common.fileutils.update_env_file")
    def test_use_with_v3_auth(self, mock_update_env_file, mock_path_exists,
                              mock_symlink, mock_remove):
        deployment_id = "593b683c-4b16-4b2b-a56b-e162bd60f10b"

        self.fake_api.deployment.get.return_value = {
            "uuid": deployment_id,
            "credentials": {
                "openstack": [{
                    "admin": {"auth_url": "http://localhost:5000/v3",
                              "username": "fake_username",
                              "password": "fake_password",
                              "tenant_name": "fake_tenant_name",
                              "endpoint": "fake_endpoint",
                              "region_name": None,
                              "user_domain_name": "fake_user_domain",
                              "project_domain_name": "fake_project_domain"}}]}}

        with mock.patch("rally.cli.commands.deployment.open", mock.mock_open(),
                        create=True) as mock_file:
            self.deployment.use(self.fake_api, deployment_id)
            self.assertEqual(3, mock_path_exists.call_count)
            mock_update_env_file.assert_has_calls([
                mock.call(os.path.expanduser("~/.rally/globals"),
                          "RALLY_DEPLOYMENT", "%s\n" % deployment_id),
                mock.call(os.path.expanduser("~/.rally/globals"),
                          "RALLY_ENV", "%s\n" % deployment_id)
            ])
            mock_file.return_value.write.assert_any_call(
                "export OS_ENDPOINT='fake_endpoint'\n")
            mock_file.return_value.write.assert_any_call(
                "export OS_AUTH_URL='http://localhost:5000/v3'\n"
                "export OS_USERNAME='fake_username'\n"
                "export OS_PASSWORD='fake_password'\n"
                "export OS_TENANT_NAME='fake_tenant_name'\n"
                "export OS_PROJECT_NAME='fake_tenant_name'\n")
            mock_file.return_value.write.assert_any_call(
                "export OS_IDENTITY_API_VERSION=3\n"
                "export OS_USER_DOMAIN_NAME='fake_user_domain'\n"
                "export OS_PROJECT_DOMAIN_NAME='fake_project_domain'\n")
            mock_symlink.assert_called_once_with(
                os.path.expanduser("~/.rally/openrc-%s" % deployment_id),
                os.path.expanduser("~/.rally/openrc"))
            mock_remove.assert_called_once_with(os.path.expanduser(
                "~/.rally/openrc"))

    @mock.patch("rally.cli.commands.deployment.DeploymentCommands."
                "_update_openrc_deployment_file")
    @mock.patch("rally.common.fileutils.update_globals_file")
    def test_use_by_name(self, mock_update_globals_file,
                         mock__update_openrc_deployment_file):
        fake_credentials = {"admin": "foo_admin", "users": ["foo_user"]}
        fake_deployment = {"uuid": "fake_uuid",
                           "credentials": {"openstack": [fake_credentials]}}
        self.fake_api.deployment.list.return_value = [fake_deployment]
        self.fake_api.deployment.get.return_value = fake_deployment
        status = self.deployment.use(self.fake_api, deployment="fake_name")
        self.assertIsNone(status)
        self.fake_api.deployment.get.assert_called_once_with(
            deployment="fake_name")
        mock_update_globals_file.assert_has_calls([
            mock.call(envutils.ENV_DEPLOYMENT, "fake_uuid"),
            mock.call(envutils.ENV_ENV, "fake_uuid")
        ])
        mock__update_openrc_deployment_file.assert_called_once_with(
            "fake_uuid", "foo_admin")

    def test_deployment_not_found(self):
        deployment_id = "e87e4dca-b515-4477-888d-5f6103f13b42"
        exc = exceptions.DBRecordNotFound(criteria="uuid: %s" % deployment_id,
                                          table="deployments")
        self.fake_api.deployment.get.side_effect = exc
        self.assertEqual(1, self.deployment.use(self.fake_api, deployment_id))

    @mock.patch("rally.cli.commands.deployment.logging.is_debug",
                return_value=False)
    @mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_deployment_check(self, mock_stdout, mock_is_debug):
        deployment_uuid = "some"
        # OrderedDict is used to predict the order of platfrom in output
        self.fake_api.deployment.check.return_value = collections.OrderedDict([
            ("openstack", [{"services": [
                {"name": "nova", "type": "compute"},
                {"name": "keystone", "type": "identity"},
                {"name": "cinder", "type": "volume"}]}]),
            ("docker", [{"admin_error": {"etype": "ProviderError",
                                         "msg": "No money - no funny!",
                                         "trace": "file1\nline1"},
                        "services": []}]),
            ("something", [{"services": [
                {"name": "foo", "type": "bar", "version": "777"},
                {"name": "xxx", "type": "yyy", "version": "777",
                 "status": "Failed", "description": "Fake service"}]},
                {"services": [], "user_error":
                    {"etype": "ProviderError",
                     "msg": "No money - no funny!",
                     "trace": "file1\nline1"}}
            ])])

        origin_print_list = cliutils.print_list

        def print_list(*args, **kwargs):
            kwargs["out"] = mock_stdout
            return origin_print_list(*args, **kwargs)

        with mock.patch.object(deployment.cliutils, "print_list",
                               new=print_list):
            self.assertEqual(
                1, self.deployment.check(self.fake_api, deployment_uuid))

        self.assertEqual(
            "-----------------------------------------------------------------"
            "---------------\nPlatform openstack:\n"
            "-----------------------------------------------------------------"
            "---------------\n\nAvailable services:\n"
            "+----------+--------------+-----------+\n"
            "| Service  | Service Type | Status    |\n"
            "+----------+--------------+-----------+\n"
            "| cinder   | volume       | Available |\n"
            "| keystone | identity     | Available |\n"
            "| nova     | compute      | Available |\n"
            "+----------+--------------+-----------+\n\n\n"
            "-----------------------------------------------------------------"
            "---------------\nPlatform docker:\n"
            "-----------------------------------------------------------------"
            "---------------\n\n"
            "Error while checking admin credentials:\n"
            "\tProviderError: No money - no funny!\n\n\n"
            "-----------------------------------------------------------------"
            "---------------\nPlatform something #1:\n"
            "-----------------------------------------------------------------"
            "---------------\n\nAvailable services:\n"
            "+---------+--------------+-----------+---------+--------------+\n"
            "| Service | Service Type | Status    | Version | Description  |\n"
            "+---------+--------------+-----------+---------+--------------+\n"
            "| foo     | bar          | Available | 777     |              |\n"
            "| xxx     | yyy          | Failed    | 777     | Fake service |\n"
            "+---------+--------------+-----------+---------+--------------+\n"
            "\n\n-------------------------------------------------------------"
            "-------------------\nPlatform something #2:\n"
            "-----------------------------------------------------------------"
            "---------------\n\n"
            "Error while checking users credentials:\n"
            "\tProviderError: No money - no funny!",
            mock_stdout.getvalue().strip())

    @mock.patch("rally.cli.commands.deployment.logging.is_debug",
                return_value=True)
    @mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_deployment_check_is_debug_turned_on(self, mock_stdout,
                                                 mock_is_debug):
        deployment_uuid = "some"
        self.fake_api.deployment.check.return_value = {
            "openstack": [{"services": [], "admin_error": {
                "etype": "KeystoneError",
                "msg": "connection refused",
                "trace": "file1\n\tline1\n\n"
                         "KeystoneError: connection refused"}}]
        }

        self.assertEqual(
            1, self.deployment.check(self.fake_api, deployment_uuid))

        self.assertEqual(
            "-----------------------------------------------------------------"
            "---------------\nPlatform openstack:\n"
            "-----------------------------------------------------------------"
            "---------------\n\n"
            "Error while checking admin credentials:\n"
            "file1\n\tline1\n\n"
            "KeystoneError: connection refused",
            mock_stdout.getvalue().strip())
